import sys, numpy, copy

from PyQt5.QtWidgets import QApplication, QMessageBox

from oasys.widgets import widget
from oasys.widgets.exchange import DataExchangeObject

from orangewidget import  gui

class SpectrumReader(widget.OWWidget):

    name = "Spectrum Reader"
    description = "Tools: Spectrum Reader"
    icon = "icons/beam_file_reader.png"
    maintainer = "Luca Rebuffi"
    maintainer_email = "lrebuffi(@at@)anl.gov"
    priority = 5
    category = "User Defined"
    keywords = ["data", "file", "load", "read"]

    outputs = [{"name":"ExchangeData",
                "type":DataExchangeObject,
                "doc":"ExchangeData",
                "id":"beam"}]

    want_main_area = 0
    want_control_area = 1

    def __init__(self):

         self.setFixedWidth(300)
         self.setFixedHeight(100)

         gui.separator(self.controlArea, height=40)
         gui.button(self.controlArea, self, "Read Spectrum File (autobinning.dat)", callback=self.read_spectrum_file)
         gui.rubber(self.controlArea)

    def read_spectrum_file(self):
        try:
            data = numpy.loadtxt("autobinning.dat", skiprows=1)

            calculated_data = DataExchangeObject(program_name="SRW", widget_name="UNDULATOR_SPECTRUM")
            calculated_data.add_content("srw_data", data)

            self.send("ExchangeData", calculated_data)
        except Exception as e:
            QMessageBox.critical(self, "Error", str(e), QMessageBox.Ok)

            if self.IS_DEVELOP: raise e

if __name__ == "__main__":
    a = QApplication(sys.argv)
    ow = SpectrumReader()
    ow.show()
    a.exec_()
    ow.saveSettings()



__author__ = 'labx'

import xraylib
from PyQt5 import QtWidgets
from oasys.menus.menu import OMenu

from orangecontrib.aps.shadow.widgets.extension.ow_power_plot_xy import PowerPlotXY

class APSShadowToolsMenu(OMenu):
    def __init__(self):
        super().__init__(name="APS Shadow Tools")

        self.openContainer()
        self.addContainer("Cumulative Plotting")
        self.addSubMenu("Enable all the Power Plot XY widgets")
        self.addSubMenu("Disable all the Power Plot XY widgets")
        self.addSeparator()
        self.addSubMenu("Select Plotting \"Yes\" in all the Power Plot XY widgets")
        self.addSubMenu("Select Plotting \"No\" in all the Power Plot XY widgets")
        self.addSeparator()
        self.addSubMenu("Clear all the cumulated plots in Power Plot XY widgets")
        self.closeContainer()

    def executeAction_1(self, action):
        try:
            for link in self.canvas_main_window.current_document().scheme().links:
                if not link.enabled:
                    widget = self.canvas_main_window.current_document().scheme().widget_for_node(link.sink_node)

                    if isinstance(widget, PowerPlotXY): link.set_enabled(True)
        except Exception as exception:
            QtWidgets.QMessageBox.critical(None, "Error",
                exception.args[0],
                QtWidgets.QMessageBox.Ok)

    def executeAction_2(self, action):
        try:
            for link in self.canvas_main_window.current_document().scheme().links:
                if link.enabled:
                    widget = self.canvas_main_window.current_document().scheme().widget_for_node(link.sink_node)

                    if isinstance(widget, PowerPlotXY): link.set_enabled(False)
        except Exception as exception:
            QtWidgets.QMessageBox.critical(None, "Error",
                exception.args[0],
                QtWidgets.QMessageBox.Ok)

    def executeAction_3(self, action):
        try:
            for node in self.canvas_main_window.current_document().scheme().nodes:
                widget = self.canvas_main_window.current_document().scheme().widget_for_node(node)

                if isinstance(widget, PowerPlotXY): widget.view_type = 1
        except Exception as exception:
            QtWidgets.QMessageBox.critical(None, "Error",
                exception.args[0],
                QtWidgets.QMessageBox.Ok)

    def executeAction_4(self, action):
        try:
            for node in self.canvas_main_window.current_document().scheme().nodes:
                widget = self.canvas_main_window.current_document().scheme().widget_for_node(node)

                if isinstance(widget, PowerPlotXY): widget.view_type = 0
        except Exception as exception:
            QtWidgets.QMessageBox.critical(None, "Error",
                exception.args[0],
                QtWidgets.QMessageBox.Ok)


    def executeAction_5(self, action):
        try:
            for node in self.canvas_main_window.current_document().scheme().nodes:
                widget = self.canvas_main_window.current_document().scheme().widget_for_node(node)

                if isinstance(widget, PowerPlotXY): widget.clearResults(interactive=False)
        except Exception as exception:
            QtWidgets.QMessageBox.critical(None, "Error",
                exception.args[0],
                QtWidgets.QMessageBox.Ok)



## @package enhanced_grid 
## @brief Contains 2D and 3D Grid containers that supports extended slicing syntax.
## 
# These classes are provided for rapid prototyping, the methods defined on them might be slow.
# @code
# grid1 = Grid3D(10, 10, 10, 0)
# grid2 = Grid3D(10, 10, 10, 1)
# grid1[0, 0, 0] = grid2[0, 0, 0]
# grid1[0, 0, 2:6:2] = grid2[0, 0, 4:5]
# grid1[0, 0, ...] = grid2[0, 0, ...]
# grid1[0, ..., 0] = grid2[..., 0, 0]
# grid1[..., ..., ...] = grid2[..., ...., ....]
# @endcode
#
# Slicing does not copy elements - an auxiliary window container is created that delegates 
# further operations to the underlying container.
# Note that assignments to slices from the same object might not behave as espected. 
# Parallel assignment also does not always work as expected.
# For example:
#@code
# grid[..., 0], grid[..., 1] = grid[..., 1], grid[..., 0]
#@endcode
# does not correctly swop two rows, but the following does:
#@code
# grid[..., 0], grid[..., 1] = grid[..., 1].clone(), grid[..., 0].clone()
#@endcode
#Strictly speaking, it is necessary only to clone the one object,
#but it is hard to remember which, so it is better to clone both (?).
# 
#@todo Implement a way to use copy containers instead of window containers

from __future__ import division
from random import randint
from math import ceil

##@brief A class that works just like a queue or a stack, except
## that a randomly selected element is returned. 
##
# This class is useful for implementing algorithms that gather 
# elements, and need to process them randomly. Something along the 
# lines of:
# 
# @code
# while not rqueue.empty():
#   #generates 3 new elements to process
#   for i in range(3): 
#     rqueue.push(process(rqueue.pop())) 
# @endcode
class RandomQueue:
	## Constructs a new empty RandomQueue
	def __init__(self):		
		## The internal list to store objects.
		self.array = []

	##Returns True if this RandomQueue is empty.	
	def empty(self):
		return len(self.array) <= 0

	## Push a new element into the RandomQueue.
	def push(self, x):
		self.array.append(x)

	## @brief Pops a randomly selected element from the queue. 
	##
	# All elements can be selected equiprobably
	def pop(self):
		n = len(self.array)

		if n <= 0:
			raise IndexError('Cannot pop from emty container!')
		elif n == 1:
			return self.array.pop()
		else:
			i = randint(0, n - 1)
			j = n - 1
			self.array[i], self.array[j] = 	self.array[j], self.array[i]

		return self.array.pop()

## @brief Class that represents a 2D array. 
##
# The following convenient syntax is supported:
# @code
# p = 2, 3 # a coordinate in the grid
# grid[p] = 5
# print grid[p] 
# print grid[2, 3] 
# @endcode
#

def signum(x):
	if x > 0:
		return 1
	elif x < 0:
		return -1
	else:
		return 0

## Truncates a point to integer coordinates.
def int_point_2d(p):
	x, y = p
	return int(x), int(y)

## Truncates a point to integer coordinates.
def int_point_3d(p):
	x, y, z = p
	return int(x), int(y), int(z)

# Every point in the sample set is represented with 
# a 1 in th grid; all other points are represented with 0.
# The returned grid is usefull for producing image data.
def points_to_grid(points, dimensions):
	grid = Grid2D(dimensions, 0)

	for point in points:
		grid[int_point_2d(point)] = 1

	return grid

## Converts a list of points to a 3D grid. 

# Every point in the sample set is represented with 
# a 1 in th grid; all other points are represented with 0.
# The returned grid is usefull for producing image data.
def points_to_grid_3d(points, dimensions):
	grid = Grid3D(dimensions, 0)

	for point in points:
		grid[int_point_3d(point)] = 1

	return grid

def make_grid_1d(width, initial_item=None):
	grid = [initial_item] * width

	return grid

## @brief Makes 2 list of lists.
def make_grid_2d(width, height, initial_item=None):
	grid = [None] * width

	for i in range(width):
		grid[i] = [None] * height

		for j in range(height):
			grid[i][j] = initial_item

	return grid

## @brief Makes 2 list of lists.
def make_grid_3d(width, height, depth, initial_item):
	grid = [None] * width

	for i in range(width):
		grid[i] = [None] * height

		for j in range(height):
			grid[i][j] = [None] * depth

			for k in range(depth):
				grid[i][j][k] = initial_item

	return grid

## @brief Returns an range that can be used to iterate over 
## the slice of the container.
#
# The following snippets are equivalent
#@code
# s = slice(3, 18, 3)
# for i in srange(s):
#	print list[i]
#@endcode
#
#@code
# for item in list[s]:
#	print item
#@endcode
def srange(s, length):
	if s == Ellipsis:
		return range(length)
	else:
		b, e, s = s.indices(length)
		return range(b, e, s)

## @brief Returns true if s is a slice or an Ellipsis.
def is_slice(s):
	return type(s) is slice or s == Ellipsis

## @brief Returns the number of elements this slice will return, provided the provided 
## primary is large enough.
def slice_len(s, length):
	if s == Ellipsis:
		return length

	b, e, s, = s.indices(length)

	tmp = int(ceil((e - b) / s))

	if tmp < 0:
		return 0
	else:
		return min(tmp, length)

	if s.stop > s.start and s.step > 0:
		return (s.stop - s.start) // s.step
	elif s.stop < s.start and s.step < 0:
		return (s.start - s.stop) // -s.step
	else:
		return 0

## @brief Returns a slice that is equivalent to the two slices combined.
##
# The following snippets are equivalent:
#@code
# list[s1][s2]
#@endcode

#@code
# list[slice_mul(s1, s2)]
#@endcode
def slice_mul(slice1, slice2, length):
	if type(slice2) is int:
		if type(slice1) is type(Ellipsis):
			return slice2
		b1, e1, s1 = slice1.indices(length)
		s2 = slice2
		if s2 < 0:
			s2 += length
		if s2 < 0:
			s2 = 0		
		return b1 + s2*s1
	elif type(slice2) is slice:
		if type(slice1) is type(Ellipsis):
			return slice2
		else:
			b1, e1, s1 = slice1.indices(length)
			b2, e2, s2 = slice2.indices(length)			
			b = b1 + b2*s1
			s = s1*s2
			e = min(b1 + e2*s1, e1)

			if e < 0 and s < 0:
				return slice(b, None, s)
			else:
				return slice(b, e,  s)



			b = slice1.start + slice2.start*slice1.step
			s = slice1.step*slice2.step
			return slice(b, min(slice1.start + slice2.stop*slice1.step, slice1.stop), s)
	elif slice2 == Ellipsis:
		return slice1

## @brief Completes this slice for a given length.
##
# The resulting slice will give the same elements for a container of the given length, but
# none of the start, stop, or step attributes will be None. If s is the Ellipses, then 
# the slice (0, length, 1) is returned.
# @deprecated
def complete_slice(s, length):
	return s

## @brief Sub-classes of this container can be used directly.
##
## A principle container will do assignment on a 1D point.
## Assignment of larger chunks is delegated to the AuxilaryContainer
## with the right dimensions.
##
## The enherritance is merely for documentation purposes.
class PrincipleContainer:
	pass

## @brief Sub-classes of this container is used as windows by a
##PrincipleContainer, and should not be used directly!
##
## An AuxiliaryContainer will do assignment a chunk with the same 
## dimensionality as itself, otherwise it delegates it to the underlying
## principle container, which will, in turn, construct the correct 
## AuxiliaryContainer to perform the assignment, or perform it if it is 
## a 1D point.
##
## The enherritance is merely for documentation purposes.
class AuxiliaryContainer:
	pass

##Abstract super class of all grid-like containers.
##These containers are static, that is, once they are created, 
##their dimensions cannot change.
#
# Children of this class must implement
# the attribute __clonetype__, which must 
# hold a callable type that can be constructed
# for clones.
# 
# Children must also implement the iterators
# cell_iter() and index_iter().
class Container:
	##
	##
	# @param dim
	#		The dimensions of this grid
	def __init__(self, dims):
		self.dims = dims

		count = 1

		for dim in dims:
			count *= dim

		self.count = count

	##Test whether two containers have the same dimensions
	##and the same items at equal indices.
	def __eq__(self, other):
		if other == None:
			return False
		if self.dims != other.dims:
			return False

		for cell1, cell2 in zip(self.cell_iter(), other.cell_iter()):
			if cell1 != cell2:
				return False
		return True

	##Equivalent to @code not (self == other) @endcode .
	def __ne__(self, other):
		return not (self == other)

	## Returns the length (1D) or width (nD) of this container.
	#
	# The length of a container is defined such that the length behaves as it would
	# for a list of lists.
	def __len__(self):
		return self.dims[0]

	## @brief Returns the minimum and maximum elements
	## of this grid as a tuple. 
	##
	#This method assumes the grid is filled.
	def min_max(self):		
		cell_iter = self.cell_iter()		
		min = max = cell_iter.next()

		for cell in cell_iter:
			if cell > max:
				max = cell
			elif cell < min:
				min = cell
		return min, max

	def copy_from(self, other):
		for index in self.index_iter():
			self[index] = other[index]

	##Makes a shallow copy of this container.
	#
	#This method constructs an instance of
	#this instance's __clonetype__. In general,
	#if this class is an AuxiliaryContainer,
	#the clone will be a PrincipleContainer
	#of the same dimension.
	def clone(self):
		new_grid = self.__clonetype__(self.dims)		
		new_grid.copy_from(self)

		return new_grid

## Class that implements __str__ and __iter__.
class Container1D (Container):
	def __init__(self, length):
		Container.__init__(self, (length,))
		self.length = length
		self.__clonetype__ = Grid1D

	def __str__(self):    
		#slow...
		glst = []

		for i in range(self.length):
			glst.append(self[i])

		return glst.__repr__()

	def __iter__(self):
		for i in range(self.length):
			yield self[i]
		raise StopIteration

	## @brief Returns the same iterator as __iter__.
	#
	# Provided so that all containers have consistent cell_iter methods.
	def cell_iter(self):
		return self.__iter__()

	## @brief Returns an iterator that iterates over a subgrid of this grid.
	##
	# The iteratir will iterate over all cells x, y in the grid	
	# such that 
	#@code
	# x0 <= x < x1
	# y0 <= y < y1
	#@endcode
	# The iterator does not iterate outside the grid.	
	def window_iter(self, x1, x0):
		for i in range(max(0, x0), min(x1, self.length)):
			yield self[i]
		raise StopIteration

	## @brief Returns an iterator that iterates over a subgrid of this grid.
	##
	# The iteratir will iterate over all cells x, y in the grid	
	# such that 
	#@code
	# x0 <= x < x1
	# y0 <= y < y1
	#@endcode
	# The iterator wraps over the grid. For example, if x is one unit too high 
	# (it is outside the grid to the right), the iterator will return first 
	# cell in that row.
	def wrapped_window_iter(self, x1, x0):
		for i in range(x0, x1):
			yield self[i % self.length]
		raise StopIteration

	## @brief Returns an iterator that iterates over all cells in the square
	## surrounding the given point. 
	##
	#The square is 2*n + 1 units.
	def square_iter(self, x, n):
		return self.window_iter(x - n, x + n + 1)

	def wrapped_square_iter(self, x, n):
		return self.wrapped_window_iter(x - n, x + n + 1)

	## @brief Returns an iterator that iterates over the indeces of this 
	## container.
	##
	# If grid is a 2 by 2 grid, then:
	# @code
	# for p in index_iter(grid):
	# 	print p
	# @endcode
	# will produce 
	# @code
	# 0, 0
	# 0, 1
	# 1, 0
	# 1, 1
	# @endcode
	# This iterator is useful for assigning elements of grids:
	# @code
	# for p in index_iter(grid):
	# 	grid[p] = random()
	# @endcode
	def index_iter(self):
		for i in range(self.length):
			yield i
		raise StopIteration

## Class that implements __str__ and __iter__.
class Container2D (Container):
	def __init__(self, width, height):
		Container.__init__(self, (width, height))
		self.width = width
		self.height = height
		self.__clonetype__ = Grid2D

	def __str__(self):    
		#slow...
		glst = []

		for i in range(self.width):
			gcol = []

			for j in range(self.height):
				gcol.append(self[i, j])

			glst.append(gcol)

		return glst.__repr__()	

	## @brief Returns an iterator that iterates over columns.
	##
	# This iterator is provided so that a Grid2D better emulates a list of 
	# lists, as in the following example:
	#@code
	# 	for col in grid:
	#		for item in col:
	#			process(item)
	#@endcode
	# Use of this iterator is discouraged - it is slow
	def __iter__(self):
		for i in range(self.width):
			yield self[i, ...]
		raise StopIteration

	## @brief Returns an iterator that iterates over all cells in the grid.
	##
	# This allows you to write:
	#@code
	# for cell in cell_iter(grid):
	#   process(cell)
	#@endcode
	def cell_iter(self):
		for i in range(self.width):
			for j in range(self.height):
				yield self[i, j]
		raise StopIteration

	## @brief Returns an iterator that iterates over a subgrid of this grid.
	##
	# The iterator will iterate over all cells x, y in the grid	
	# such that 
	#@code
	# x0 <= x < x1
	# y0 <= y < y1
	#@endcode
	#	
	def window_index_iter(self, p0, p1):
		x0, y0 = p0
		x1, y1 = p1    
		for i in range(max(0, x0), min(x1, self.width)):
			for j in range(max(0, y0), min(y1, self.height)):
				yield (i, j)
		raise StopIteration
	## @brief Returns an iterator that iterates over a subgrid of this grid.
	##
	# The iterator will iterate over all cells x, y in the grid	
	# such that 
	#@code
	# x0 <= x < x1
	# y0 <= y < y1
	#@endcode
	#	
	# The iterator wraps over the grid. For example, if x is one unit too high 
	# (it is outside the grid to the right), the iterator will return the index of the 
	# first cell in that row.
	def wrapped_window_index_iter(self, p0, p1):
		x0, y0 = p0
		x1, y1 = p1    
		for i in range(x0, x1):
			for j in range(y0, y1):
				yield (i % self.width, j % self.height)
		raise StopIteration

	## @brief Returns an iterator that iterates over a subgrid of this grid.
	##
	# The iterator will iterate over all cells x, y in the grid	
	# such that 
	#@code
	# x0 <= x < x1
	# y0 <= y < y1
	#@endcode	
	def window_iter(self, p0, p1):
		x0, y0 = p0
		x1, y1 = p1
		for i in range(max(0, x0), min(x1, self.width)):
			for j in range(max(0, y0), min(y1, self.height)):
				yield self[i, j]
		raise StopIteration	

	## @brief Returns an iterator that iterates over a subgrid of this grid.
	##
	# The iterator will iterate over all cells x, y in the grid	
	# such that 
	#@code
	# x0 <= x < x1
	# y0 <= y < y1
	#@endcode
	#	
	# The iterator wraps over the grid. For example, if x is one unit too high 
	# (it is outside the grid to the right), the iterator will return first 
	# cell in that row.
	def wrapped_window_iter(self, p0, p1):
		x0, y0 = p0
		x1, y1 = p1
		for i in range(x0, x1):
			for j in range(y0, y1):
				yield self[i % self.width, j % self.height]
		raise StopIteration	

	## @brief Returns an iterator that iterates over all cells in the square
	## surrounding the given point. 
	##
	#The square is 2*n + 1 units.
	def square_index_iter(self, p, n):
		x, y = p
		return self.window_index_iter((x - n, y - n), (x + n + 1, y + n +1))
	
	## @brief Returns an iterator that iterates over all cells in the square
	## surrounding the given point. 
	##
	#The square is 2*n + 1 units.
	# The iterator wraps over the grid. For example, if x is one unit too high 
	# (it is outside the grid to the right), the iterator will return first 
	# cell in that row.
	def wrapped_square_index_iter(self, p, n):
		x, y = p
		return self.wrapped_window_index_iter((x - n, y - n), (x + n + 1, y + n +1))


	## @brief Returns an iterator that iterates over all cells in the square
	## surrounding the given point. 
	##
	#The square is 2*n + 1 units.
	def square_iter(self, p, n):
		x, y = p
		return self.window_iter((x - n, y - n), (x + n + 1, y + n +1))

	## @brief Returns an iterator that iterates over all cells in the square
	## surrounding the given point. 
	##
	#The square is 2*n + 1 units.
	# The iterator wraps over the grid. For example, if x is one unit too high 
	# (it is outside the grid to the right), the iterator will return first 
	# cell in that row.
	def wrapped_square_iter(self, p, n):
		x, y = p
		return self.wrapped_window_iter((x - n, y - n), (x + n + 1, y + n +1))

	## @brief Returns an iterator that iterates over the indeces of this 
	## grid as tuples.
	##
	# If grid is a 2 by 2 grid, then:
	# @code
	# for p in index_iter(grid):
	# 	print p
	# @endcode
	# will produce 
	# @code
	# 0, 0
	# 0, 1
	# 1, 0
	# 1, 1
	# @endcode
	# This iterator is useful for assigning elements of grids:
	# @code
	# for p in index_iter(grid):
	# 	grid[p] = random()
	# @endcode
	def index_iter(self):
		for i in range(self.width):
			for j in range(self.height):
				yield i, j
		raise StopIteration

## Class that implements __str__ and __iter__.
class Container3D (Container):
	def __init__(self, width, height, depth):
		Container.__init__(self, (width, height, depth))
		self.width = width
		self.height = height
		self.depth = depth

		self.__clonetype__ = Grid3D

	def __str__(self):    
		#slow...
		glst = []

		for i in range(self.width):
			gcol = []
			for j in range(self.height):
				gslice = []

				for k in range(self.depth):
					gslice.append(self[i, j, k])

				gcol.append(gslice)
			glst.append(gcol)

		return glst.__repr__()

	def __iter__(self):
		for i in range(self.width):
			yield self[i, ..., ...]
		raise StopIteration

	## Returns an iterator that iterates over all cells in the grid
	def cell_iter(self):
		for i in range(self.width):
			for j in range(self.height):
				for k in range(self.depth):
					yield self[i, j, k]
		raise StopIteration

	## @brief Returns an iterator that iterates over the indeces of this 
	## grid as tuples.
	##
	# If grid is a 2 by 2 grid, then:
	# @code
	# for p in index_iter(grid):
	# 	print p
	# @endcode
	# will produce 
	# @code
	# 0, 0
	# 0, 1
	# 1, 0
	# 1, 1
	# @endcode
	# This iterator is useful for assigning elements of grids:
	# @code
	# for p in index_iter(grid):
	# 	grid[p] = random()
	# @endcode
	def index_iter(self):
		for i in range(self.width):
			for j in range(self.height):
				for k in range (self.depth):
					yield i, j, k
		raise StopIteration	

	## @brief Returns an iterator that iterates over a subgrid of this grid.
	##
	# The iterator will iterate over all cells x, y, z in the grid
	# such that 
	#@code
	# x0 <= x < x1
	# y0 <= y < y1
	# z0 <= z < z1
	#@endcode
	#
	def window_iter(self, p0, p1):
		x0, y0, z0 = p0
		x1, y1, z1 = p1

		for i in range(max(0, x0), min(x1, self.width)):
			for j in range(max(0, y0), min(y1, self.height)):
				for k in range(max(0, z0), min(z1, self.depth)):
					yield self[i, j, k]
		raise StopIteration

	## @brief Returns an iterator that iterates over a subgrid of this grid.
	##
	# The iterator will iterate over all cells x, y, z in the grid
	# such that 
	#@code
	# x0 <= x < x1
	# y0 <= y < y1
	# z0 <= z < z1
	#@endcode
	#wrapping around the edges as necessary.
	
	def wrapped_window_iter(self, p0, p1):
		x0, y0, z0 = p0
		x1, y1, z1 = p1

		for i in range(x0, x1):
			for j in range(y0, y1):
				for k in range(z0, z1):
					yield self[i % self.width, j % self.height, k % self.depth]
		raise StopIteration
	## @brief Returns an iterator that iterates over all cells in the square
	##	surrounding the given point. 
	##
	#The cube is 2*n + 1 units.
	def square_iter(self, p, n):
		x, y, z = p
		return self.window_iter(
			(x - n, y - n, z - n), 
			(x + n + 1, y + n +1, z + n +1))
	## @brief Returns an iterator that iterates over all cells in the square
	##	surrounding the given point, wrapping around as necessary.
	##
	#The cube is 2*n + 1 units.
	def wrapped_square_iter(self, p, n):
		x, y, z = p
		return self.wrapped_window_iter(
			(x - n, y - n, z - n), 
			(x + n + 1, y + n +1, z + n +1))

class GridWindow1D (Container1D, AuxiliaryContainer): #Constant y
	def __init__(self, grid, col_slice):
		self.grid = grid
		self.x = complete_slice(col_slice, grid.width)		
		Container1D.__init__(self, slice_len(self.x), grid.width)

	def __getitem__(self, x):
		new_x = slice_mul(self.x, x, self.grid.width)

		return self.grid[new_x]		

	def __setitem__(self, x, item):
		new_x = slice_mul(self.x, x, self.grid.width)

		if type(x) is int:
			self.grid[new_x] = item
		else: #slice!
			for i, item_i in zip(srange(new_x, self.grid.width), item):
				self.grid[i] = item_i

## Class that represent a 2D grid, with enhanced slicing notation.
class Grid1D (Container1D, PrincipleContainer):
	def __init__(self, dims, initial_item = None):
		(width,) = dims
		Container1D.__init__(self, width)
		self.grid = make_grid_1d(width, initial_item)
		self.width = width
	## @brief Returns an iterator that iterates over all cells in the grid.
	##
	# This allows you to write:
	#@code
	# for cell in cell_iter(grid):
	#   process(cell)
	#@endcode
	def cell_iter(self):
		for i in range(self.width):
			yield self.grid[i]
		raise StopIteration

	def __getitem__(self, x):
		if isinstance(x, int):
			return self.grid[x]
		elif is_slice(x):
			GridWindow1D(self, x)
		raise TypeError

	def __setitem__(self, x, item):		
		if type(x) is int:			
			self.grid[x] = item
		elif is_slice(x):
			g = GridWindow1D(self, x)
			g[...] = item
		else:
			raise TypeError

class GridRow2D (Container1D, AuxiliaryContainer): #Constant y
	def __init__(self, grid, col_slice, row):
		self.grid = grid
		self.x = complete_slice(col_slice, grid.width)
		self.y = row

		Container1D.__init__(self, slice_len(self.x, grid.width))

	def __getitem__(self, x):
		new_x = slice_mul(self.x, x, self.grid.width)

		return self.grid[new_x, self.y]		

	def __setitem__(self, x, item):
		new_x = slice_mul(self.x, x, self.grid.width)

		if type(x) is int:
			self.grid[new_x] = item
		else: #slice!
			for i, item_i in zip(srange(new_x, self.grid.width), item):
				self.grid[i, self.y] = item_i

class GridCol2D (Container1D, AuxiliaryContainer): #Constant x
	def __init__(self, grid, col, row_slice):
		self.grid = grid
		self.x = col
		self.y = complete_slice(row_slice, grid.height)

		Container1D.__init__(self, slice_len(self.y, grid.height))

	def __getitem__(self,  y):
		new_y = slice_mul(self.y, y, self.grid.height)

		return self.grid[self.x, new_y]

	def __setitem__(self, y, item):
		new_y = slice_mul(self.y, y, self.grid.height)

		if type(y) is int:			
			self.grid[self.x, new_y] = item
		else: #slice!
			for i, item_i in zip(srange(new_y, self.grid.height), item):
				self.grid[self.x, i] = item_i

class GridWindow2D (Container2D):
	def __init__(self, grid, x, y):
		self.grid = grid
		self.x = complete_slice(x, grid.width)
		self.y = complete_slice(y, grid.height)

		Container2D.__init__(self, slice_len(self.x, grid.width), slice_len(self.y, grid.height))

	def __getitem__(self, p):
		if isinstance(p, int):
			return self[p, ...]			
		x, y = p			

		new_x = slice_mul(self.x, x, self.grid.width)
		new_y = slice_mul(self.y, y, self.grid.height)
		return self.grid[new_x, new_y]

	def __setitem__(self, p, item):
		if isinstance(p, int):
			self[p, ...] = item
		x, y = p
		new_x = slice_mul(self.x, x, self.grid.width)
		new_y = slice_mul(self.y, y, self.grid.height)

		if type(x) is int or type(y) is int:
			#delegate!			
			self.grid[new_x, new_y] = item
		else: #slice!
			for i, item_i in zip(srange(new_x, self.grid.width), item):
				for j, item_j in zip(srange(new_y, self.grid.height), item_i):
					self.grid[i, j] = item_j

	def __repr__(self):    
		#slow...
		glst = []

		for i in range(slice_len(self.x, self.grid.width)):
			gcol = []
			for j in range(slice_len(self.y, self.grid.height)):
				gcol.append(self[i, j])
			glst.append(gcol)

		return glst.__repr__()			

## Class that represent a 2D grid, with enhanced slicing notation.
class Grid2D (Container2D, PrincipleContainer):

	def __init__(self, dims, initial_item = None):
		(width, height) = dims
		Container2D.__init__(self, width, height)
		self.grid = make_grid_2d(width, height, initial_item)
	## @brief Returns an iterator that iterates over all cells in the grid.
	##
	# This allows you to write:
	#@code
	# for cell in cell_iter(grid):
	#   process(cell)
	#@endcode
	def cell_iter(self):
		for i in range(self.width):
			for j in range(self.height):
				yield self.grid[i][j]
		raise StopIteration

	def __getitem__(self, p):
		if isinstance(p, int):
			return self[p, ...]
		x, y = p
		if isinstance(x, int):
			if isinstance(y, int):
				return self.grid[x][y]
			elif is_slice(y):
				return GridCol2D(self, x, y)
		elif is_slice(x):
			if isinstance(y, int):
				return GridRow2D(self, x, y)
			elif is_slice(y):
				return GridWindow2D(self, x, y)

		raise TypeError

	def __setitem__(self, p, item):		
		x, y = p
		if type(x) is int:
			if type(y) is int:
				self.grid[x][y] = item
			elif is_slice(y):
				g = GridCol2D(self, x, y)
				g[...] = item
		elif is_slice(x):
			if type(y) is int:
				g = GridRow2D(self, x, y)
				g[...] = item
			elif is_slice(y):
				g = GridWindow2D(self, x, y)
				g[..., ...] = item
		else:
			raise TypeError

class GridBar3D (Container1D, AuxiliaryContainer): #constant x, y
	def __init__(self, grid, x, y, z):
		self.grid = grid
		self.x = x
		self.y = y
		self.z = complete_slice(z, grid.depth)

		Container1D.__init__(self, slice_len(self.z, grid.depth))

	def __getitem__(self,  z):
		new_z = slice_mul(self.z, z, self.grid.depth)		
		return self.grid[self.x, self.y, new_z]

	def __setitem__(self, z, item):
		new_z = slice_mul(self.z, z, self.grid.depth)

		if type(z) is int:
			self.grid[new_z] = item
		else: #slice!
			for i, item_i in zip(srange(new_z, self.grid.depth), item):
				self.grid[self.x, self.y, i] = item_i	

class GridCol3D (Container1D, AuxiliaryContainer): #constant x, z
	def __init__(self, grid, x, y, z):
		self.grid = grid
		self.x = x
		self.y = complete_slice(y, grid.height)
		self.z = z

		Container1D.__init__(self, slice_len(self.y, grid.height))

	def __getitem__(self,  y):
		new_y = slice_mul(self.y, y, self.grid.height)		
		return self.grid[self.x, new_y, self.z]
	def __setitem__(self, y, item):
		new_y = slice_mul(self.y, y, self.grid.height)

		if type(y) is int:
			self.grid[new_y] = item
		else: #slice!
			for i, item_i in zip(srange(new_y, self.grid.height), item):
				self.grid[self.x, i, self.z] = item_i	

class GridRow3D (Container1D, AuxiliaryContainer): #constant y, z
	def __init__(self, grid, x, y, z):
		self.grid = grid
		self.x = complete_slice(x, grid.width)
		self.y = y
		self.z = z

		Container1D.__init__(self, slice_len(self.x, grid.width))
	def __getitem__(self,  x):
		new_x = slice_mul(self.x, x, self.grid.width)		
		return self.grid[new_x, self.y, self.z]
	def __setitem__(self, x, item):
		new_x = slice_mul(self.x, x, self.grid.width)

		if type(x) is int:
			self.grid[new_x] = item
		else: #slice!
			for i, item_i in zip(srange(new_x, self.grid.width), item):
				self.grid[i, self.y, self.y] = item_i

class GridSliceXY (Container2D, AuxiliaryContainer): #constant z
	def __init__(self, grid, x, y, z):
		self.grid = grid
		self.x = complete_slice(x, grid.width)
		self.y = complete_slice(y, grid.height)
		self.z = z		
		Container2D.__init__(self, slice_len(self.x, grid.width), slice_len(self.y, grid.height))
	def __getitem__(self,  p):
		if isinstance(p, int):
			return self[p, ...]			
		x, y = p	
		new_x = slice_mul(self.x, x, self.grid.width)		
		new_y = slice_mul(self.y, y, self.grid.height)
		return self.grid[new_x, new_y, self.z]

	def __setitem__(self, p, item):
		if isinstance(p, int):
			self[p, ...] = item
		x, y = p	
		new_x = slice_mul(self.x, x, self.grid.width)
		new_y = slice_mul(self.y, y, self.grid.height)

		if type(x) is int or type(y) is int:
			#delegate!
			self.grid[new_x, new_y, self.z] = item
		else: #slice!
			for i, item_i in zip(srange(new_x, self.grid.width), item):
				for j, item_j in zip(srange(new_y, self.grid.height), item_i):
					self.grid[i, j, self.z] = item_j

class GridSliceXZ (Container2D, AuxiliaryContainer): #constant Y
	def __init__(self, grid, x, y, z):
		self.grid = grid
		self.x = complete_slice(x, grid.width)
		self.y = y
		self.z = complete_slice(z, grid.depth)

		Container2D.__init__(self, slice_len(self.x, grid.width), slice_len(self.z, grid.depth))
	def __getitem__(self,  p):
		if isinstance(p, int):
			return self[p, ...]			
		x, z = p			
		new_x = slice_mul(self.x, x, self.grid.width)		
		new_z = slice_mul(self.z, z, self.grid.depth)
		return self.grid[new_x, self.y, new_z]

	def __setitem__(self, p, item):
		if isinstance(p, int):
			return self[p, ...]			
		x, z = p	
		new_x = slice_mul(self.x, x, self.grid.width)
		new_z = slice_mul(self.z, z, self.grid.depth)

		if type(x) is int or type(z) is int:
			#delegate!
			self.grid[new_x, self.y, new_z] = item
		else: #slice!
			for i, item_i in zip(srange(new_x, self.grid.width), item):
				for j, item_j in zip(srange(new_z, self.grid.depth), item_i):
					self.grid[i, self.y, j] = item_j		

class GridSliceYZ (Container2D, AuxiliaryContainer): #constant x
	def __init__(self, grid, x, y, z):
		self.grid = grid
		self.x = x
		self.y = complete_slice(y, grid.height)
		self.z = complete_slice(z, grid.depth)

		Container2D.__init__(self, slice_len(self.y, grid.height), slice_len(self.z, grid.depth))
	def __getitem__(self,  p):
		if isinstance(p, int):
			return self[p, ...]			
		y, z = p	
		new_y = slice_mul(self.y, y, self.grid.height)		
		new_z = slice_mul(self.z, z, self.grid.depth)
		return self.grid[self.x, new_y, new_z]

	def __setitem__(self, p, item):
		if isinstance(p, int):
			self[p, ...] = item
		y, z = p	
		new_y = slice_mul(self.y, y, self.grid.height)
		new_z = slice_mul(self.z, z, self.grid.depth)

		if type(y) is int or type(z) is int:
			#delegate!
			self.grid[self.x, new_y, new_z] = item
		else: #slice!
			for i, item_i in zip(srange(new_y), item):
				for j, item_j in zip(srange(new_z), item_i):
					self.grid[self.x, i, j] = item_j

class GridWindow3D (Container3D, AuxiliaryContainer):
	def __init__(self, grid, x, y, z):
		self.grid = grid
		self.x = complete_slice(x, grid.width)
		self.y = complete_slice(y, grid.height)
		self.z = complete_slice(z, grid.height)	

		Container3D.__init__(self, slice_len(self.x, grid.width), slice_len(self.y, grid.height), slice_len(self.z, grid.depth))
	def __getitem__(self, p):
		if isinstance(p, int):
			return self[p, ..., ...]			
		x, y, z = p	
		new_x = slice_mul(self.x, x, self.grid.width)
		new_y = slice_mul(self.y, y, self.grid.height)
		new_z = slice_mul(self.z, z, self.grid.depth)

		return self.grid[new_x, new_y, new_z]

	def __setitem__(self, p, item):
		if isinstance(p, int):
			self[p, ..., ...] = item
		x, y, z = p	
		new_x = slice_mul(self.x, x, self.grid.width)
		new_y = slice_mul(self.y, y, self.grid.height)
		new_z = slice_mul(self.z, z, self.grid.depth)

		if type(x) is int or type(y) is int or type(z) is int:
			#delegate!
			self.grid[new_x, new_y, self.z] = item
		else: #slice!
			for i, item_i in zip(srange(new_x, self.grid.width), item):
				for j, item_j in zip(srange(new_y, self.grid.height), item_i):
					for k, item_k in zip(srange(new_z, self.grid.depth), item_j):
						self.grid[i, j, k] = item_k

## Class that represent a 3D grid, with enhanced slicing notation.
class Grid3D (Container3D, PrincipleContainer):
	def __init__(self, dims, initial_item = None):
		(width, height, depth) = dims
		Container3D.__init__(self, width, height, depth)
		self.grid = make_grid_3d(width, height, depth, initial_item)

	def __getitem__(self, p):
		if isinstance(p, int):
			return self[p, ..., ...]
		x, y, z = p
		if type(x) is int:
			if type(y) is int:
				if type(z) is int:
					return self.grid[x][y][z]
				elif is_slice(z):
					return GridBar3D(self, x, y, z)
			elif is_slice(y):
				if type(z) is int:
					return GridCol3D(self, x, y, z)
				elif is_slice(z):
					return GridSliceYZ(self, x, y, z)
		elif is_slice(x):
			if type(y) is int:
				if type(z) is int:
					return GridRow3D(self, x, y, z)
				elif is_slice(z):
					return GridSliceXZ(self, x, y, z)
			elif is_slice(y):
				if type(z) is int:
					return GridSliceXY(self, x, y, z)
				elif is_slice(z):
					return GridWindow3D(self, x, y, z)

		raise TypeError

	def __setitem__(self, p, item):
		(x, y, z) = p
		if type(x) is int:
			if type(y) is int:
				if type(z) is int:
					self.grid[x][y][z] = item
				elif is_slice(z):
					g = GridBar3D(self, x, y, z)
					g[...] = item
			elif is_slice(y):
				if type(z) is int:
					g = GridCol3D(self, x, y, z)
					g[...] = item
				elif is_slice(z):
					g = GridSliceYZ(self, x, y, z)
					g[..., ...] = item
		elif is_slice(x):
			if type(y) is int:
				if type(z) is int:
					g = GridRow3D(self, x, y, z)
					g[...] = item
				elif is_slice(z):
					g = GridSliceXZ(self, x, y, z)
					g[..., ...] = item
			elif is_slice(y):
				if type(z) is int:
					g = GridSliceXY(self, x, y, z)
					g[..., ...] = item
				elif is_slice(z):
					g = GridWindow3D(self, x, y, z)
					g[..., ..., ...] = item
		else:
			raise TypeError

class ListGrid3D (Grid3D):
	def __init__(self, dims):
		(width, height, depth) = dims
		Grid3D.__init__(self, (width, height, depth))

		for index in self.index_iter():
			self[index] = []

	## @brief Sets the item at x, y.
	##
	# Use a tuplet (x, y) to access the item.
	def additem(self, p, b):
		(x, y, z) = p
		self[x, y, z].append(b)

class ListGrid2D (Grid2D):
	def __init__(self, dims):
		(width, height) = dims
		Grid2D.__init__(self, (width, height))

		for index in self.index_iter():
			self[index] = []

	## @brief Sets the item at x, y.
	##
	# Use a tuplet (x, y) to access the item.
	def additem(self, p, b):
		(x, y) = p
		self[x, y].append(b)




import os, numpy

from orangewidget import gui
from oasys.widgets import gui as oasysgui

from oasys.widgets import widget

from PyQt5.QtGui import QPalette, QColor, QFont
from PyQt5.QtWidgets import QApplication, QMessageBox
from PyQt5.QtCore import QRect

from oasys.util.oasys_objects import OasysPreProcessorData, OasysSurfaceData

from Shadow import ShadowTools as ST
from orangecontrib.shadow.util.shadow_objects import ShadowPreProcessorData


class OWMultipleOasysDataConverter(widget.OWWidget):
    name = "Multiple Oasys Surface Data Converter"
    id = "oasysDataConverter"
    description = "Multiple Oasys Surface Data Converter"
    icon = "icons/multiple_oasys_data_converter.png"
    priority = 0
    category = ""
    keywords = ["wise", "gaussian"]

    inputs = [("Oasys PreProcessorData", OasysPreProcessorData, "set_input")]

    outputs = [{"name": "PreProcessor_Data",
                "type": ShadowPreProcessorData,
                "doc": "PreProcessor Data",
                "id": "PreProcessor_Data"},
               {"name":"Files",
                "type":list,
                "doc":"Files",
                "id":"Files"}]

    CONTROL_AREA_WIDTH = 605

    want_main_area = 0

    oasys_data = None

    def __init__(self):
        super().__init__()

        geom = QApplication.desktop().availableGeometry()
        self.setGeometry(QRect(round(geom.width()*0.05),
                               round(geom.height()*0.05),
                               round(min(geom.width()*0.98, self.CONTROL_AREA_WIDTH+10)),
                               round(min(geom.height()*0.95, 100))))

        self.setMaximumHeight(self.geometry().height())
        self.setMaximumWidth(self.geometry().width())

        self.controlArea.setFixedWidth(self.CONTROL_AREA_WIDTH)


        label = gui.label(self.controlArea, self, "From Multiple Oasys Surfaces To Shadow Surfaces")
        font = QFont(label.font())
        font.setBold(True)
        font.setItalic(True)
        font.setPixelSize(14)
        label.setFont(font)
        palette = QPalette(label.palette()) # make a copy of the palette
        palette.setColor(QPalette.Foreground, QColor('Dark Blue'))
        label.setPalette(palette) # assign new palette

        gui.separator(self.controlArea, 10)

        gui.button(self.controlArea, self, "Convert", callback=self.convert_surfaces, height=45)

    def set_input(self, input_data):
        self.setStatusMessage("")

        if not input_data is None:
            self.oasys_data = input_data

            self.convert_surfaces()

    def convert_surfaces(self):
        if not self.oasys_data is None:
            try:
                if isinstance(self.oasys_data, OasysPreProcessorData):
                    error_profile_data = self.oasys_data.error_profile_data
                    surface_data = error_profile_data.surface_data

                    error_profile_data_files = []

                    for xx, yy, zz, error_profile_data_file in zip(surface_data.xx,
                                                                   surface_data.yy,
                                                                   surface_data.zz,
                                                                   surface_data.surface_data_file):

                        filename, file_extension = os.path.splitext(error_profile_data_file)

                        if (file_extension==".hd5" or file_extension==".hdf5" or file_extension==".hdf"):
                            error_profile_data_file = filename + "_shadow.dat"

                        ST.write_shadow_surface(zz/self.workspace_units_to_m,
                                                numpy.round(xx/self.workspace_units_to_m, 6),
                                                numpy.round(yy/self.workspace_units_to_m, 6),
                                                error_profile_data_file)

                        error_profile_data_files.append(error_profile_data_file)

                    self.send("PreProcessor_Data", ShadowPreProcessorData(error_profile_data_file=error_profile_data_files,
                                                                          error_profile_x_dim=error_profile_data.error_profile_x_dim/self.workspace_units_to_m,
                                                                          error_profile_y_dim=error_profile_data.error_profile_y_dim/self.workspace_units_to_m))
                    self.send("Files", error_profile_data_files)
            except Exception as exception:
                QMessageBox.critical(self, "Error", str(exception), QMessageBox.Ok)

                if self.IS_DEVELOP: raise exception



import sys, numpy, os

from PyQt5 import QtGui
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QMessageBox, QLabel, QSizePolicy
from PyQt5.QtGui import QPixmap

import orangecanvas.resources as resources

from orangewidget import gui
from orangewidget.widget import OWAction

from oasys.widgets.exchange import DataExchangeObject
from oasys.widgets import gui as oasysgui

from orangecontrib.shadow.util.shadow_objects import ShadowBeam
from orangecontrib.shadow.util.shadow_util import ShadowCongruence
from orangecontrib.shadow.widgets.gui.ow_automatic_element import AutomaticElement

class FluxCalculator(AutomaticElement):

    name = "Flux Calculator"
    description = "Tools: Flux Calculator"
    icon = "icons/flux.png"
    maintainer = "Luca Rebuffi"
    maintainer_email = "lrebuffi(@at@)anl.gov"
    priority = 10
    category = "User Defined"
    keywords = ["data", "file", "load", "read"]

    inputs = [("Shadow Beam", ShadowBeam, "setBeam"),
              ("Spectrum Data", DataExchangeObject, "setSpectrumData")]

    outputs = [{"name":"Beam",
                "type":ShadowBeam,
                "doc":"Shadow Beam",
                "id":"beam"}]

    want_main_area = 0
    want_control_area = 1

    input_beam     = None
    input_spectrum = None
    flux_index = -1

    usage_path = os.path.join(resources.package_dirname("orangecontrib.aps.shadow.widgets.extension"), "misc", "flux_calculator.png")

    def __init__(self):
        super(FluxCalculator, self).__init__()

        self.runaction = OWAction("Calculate Flux", self)
        self.runaction.triggered.connect(self.calculate_flux)
        self.addAction(self.runaction)

        self.setMaximumWidth(self.CONTROL_AREA_WIDTH+10)
        self.setMaximumHeight(580)

        box0 = gui.widgetBox(self.controlArea, "", orientation="horizontal")
        gui.button(box0, self, "Calculate Flux", callback=self.calculate_flux, height=45)

        tabs_setting = oasysgui.tabWidget(self.controlArea)
        tabs_setting.setFixedHeight(440)
        tabs_setting.setFixedWidth(self.CONTROL_AREA_WIDTH-8)

        tab_out = oasysgui.createTabPage(tabs_setting, "Flux Calculation Results")
        tab_usa = oasysgui.createTabPage(tabs_setting, "Use of the Widget")
        tab_usa.setStyleSheet("background-color: white;")

        self.text = oasysgui.textArea(width=self.CONTROL_AREA_WIDTH-22, height=400)

        tab_out.layout().addWidget(self.text)

        usage_box = oasysgui.widgetBox(tab_usa, "", addSpace=True, orientation="horizontal")

        label = QLabel("")
        label.setAlignment(Qt.AlignCenter)
        label.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        label.setPixmap(QPixmap(self.usage_path))

        usage_box.layout().addWidget(label)





        gui.rubber(self.controlArea)

    def setBeam(self, beam):
        try:
            if ShadowCongruence.checkEmptyBeam(beam):
                if ShadowCongruence.checkGoodBeam(beam):
                    self.input_beam = beam

                    if self.is_automatic_run: self.calculate_flux()
        except Exception as exception:
            QMessageBox.critical(self, "Error", str(exception), QMessageBox.Ok)

            if self.IS_DEVELOP: raise exception

    def setSpectrumData(self, data):
        if not data is None:
            try:
                if data.get_program_name() == "XOPPY":
                    if data.get_widget_name() == "UNDULATOR_FLUX" or data.get_widget_name() == "XWIGGLER" or data.get_widget_name() == "WS":
                        self.flux_index = 1
                    elif data.get_widget_name() == "BM":
                        self.flux_index = 5
                    else:
                        raise Exception("Connect to one of the following XOPPY widgets: Undulator Spectrum, BM, XWIGGLER, WS")

                    self.input_spectrum = data.get_content('xoppy_data')
                elif data.get_program_name() == "SRW":
                    if data.get_widget_name() == "UNDULATOR_SPECTRUM":
                        self.flux_index = 1
                    else:
                        raise Exception("Connect to one of the following SRW widgets: Undulator Spectrum")

                    self.input_spectrum = data.get_content('srw_data')
                else:
                    raise ValueError("Widget accept data from the following Add-ons: XOPPY, SRW")

                if self.is_automatic_run: self.calculate_flux()
            except Exception as exception:
                QMessageBox.critical(self, "Error", str(exception), QMessageBox.Ok)

                if self.IS_DEVELOP: raise exception

    def calculate_flux(self):
        if not self.input_beam is None and not self.input_spectrum is None:
            try:
                flux_factor, resolving_power, energy, ttext = calculate_flux_factor_and_resolving_power(self.input_beam)

                total_text = ttext

                flux_at_sample, ttext = calculate_flux_at_sample(self.input_spectrum, self.flux_index, flux_factor, energy)

                total_text += "\n" + ttext

                total_text += "\n\n ---> Flux at Image Plane : %g"%flux_at_sample + " ph/s"
                total_text += "\n ---> Resolving Power: %g"%resolving_power

                self.text.clear()
                self.text.setText(total_text)

                self.send("Beam", self.input_beam)
            except Exception as exception:
                QMessageBox.critical(self, "Error", str(exception), QMessageBox.Ok)

                if self.IS_DEVELOP: raise exception

def calculate_flux_factor_and_resolving_power(beam):
    ticket = beam._beam.histo1(11, nbins=2, nolost=1)

    energy_min = ticket['xrange'][0]
    energy_max = ticket['xrange'][-1]

    Denergy_source = numpy.abs(energy_max - energy_min)
    energy = numpy.average([energy_min, energy_max])

    if Denergy_source == 0.0:
        raise ValueError("This calculation is not possibile for a single energy value")

    ticket = beam._beam.histo1(11, nbins=200, nolost=1, ref=23)

    initial_intensity = len(beam._beam.rays)
    final_intensity = ticket['intensity']
    efficiency = final_intensity/initial_intensity
    bandwidth = ticket['fwhm']

    if bandwidth == 0.0:
        raise ValueError("Bandwidth is 0.0: calculation not possible")

    resolving_power = energy/bandwidth

    if Denergy_source < 4*bandwidth:
        raise ValueError("Source \u0394E (" + str(round(Denergy_source, 2)) + " eV) should be at least 4 times bigger than the bandwidth (" + str(round(bandwidth, 3)) + " eV)")

    text = "\n# SOURCE ---------\n"
    text += "\n Source Central Energy: %g"%round(energy, 2) + " eV"
    text += "\n Source Energy Range  : %g - %g"%(round(energy_min, 2), round(energy_max, 2)) + " eV"
    text += "\n Source \u0394E: %g"%round(Denergy_source, 2) + " eV"

    text += "\n\n# BEAMLINE ---------\n"
    text += "\n Shadow Intensity (Initial): %g"%initial_intensity
    text += "\n Shadow Intensity (Final)  : %g"%final_intensity
    text += "\n"
    text += "\n Efficiency: %g"%round(100*efficiency, 3) + "%"
    text += "\n Bandwidth (at the Image Plane): %g"%round(bandwidth, 3) + " eV"

    beamline_bandwidth = Denergy_source * efficiency

    flux_factor = beamline_bandwidth / (1e-3*energy)

    return flux_factor, resolving_power, energy, text

def calculate_flux_at_sample(spectrum, flux_index, flux_factor, energy):
    index_up = numpy.where(spectrum[:, 0] >= energy)
    index_down = numpy.where(spectrum[:, 0] < energy)

    flux_up = spectrum[index_up, flux_index][0, 0]
    flux_down = spectrum[index_down, flux_index][0, -1]

    interpolated_flux = (flux_up + flux_down)/2

    text = "\n# FLUX INTERPOLATION ---------\n"
    text += "\n Energy range: %g - %g"%(spectrum[index_down, 0][0, -1], spectrum[index_up, 0][0, 0]) + " eV"
    text += "\n Spectral Flux Density: %g"%interpolated_flux + " ph/s/0.1%bw"

    return interpolated_flux*flux_factor, text


if __name__ == "__main__":
    a = QtGui.QApplication(sys.argv)
    ow = FluxCalculator()
    ow.show()
    a.exec_()
    ow.saveSettings()





import sys, numpy, copy

from oasys.widgets import widget
from oasys.util.oasys_util import TriggerIn

from orangewidget import gui

from PyQt5 import QtGui

from orangecontrib.shadow.util.shadow_objects import ShadowBeam
from orangecontrib.shadow.util.shadow_util import ShadowCongruence

class BeamCleaner(widget.OWWidget):

    name = "Beam Cleaner"
    description = "Tools: Beam Cleaner"
    icon = "icons/clean.png"
    maintainer = "Luca Rebuffi"
    maintainer_email = "luca.rebuffi(@at@)elettra.eu"
    priority = 30
    category = "User Defined"
    keywords = ["data", "file", "load", "read"]

    inputs = [("Input Beam", ShadowBeam, "setBeam")]

    outputs = [{"name":"Beam",
                "type":ShadowBeam,
                "doc":"Shadow Beam",
                "id":"beam"},
               {"name":"Trigger",
                "type": TriggerIn,
                "doc":"Feedback signal to start a new beam simulation",
                "id":"Trigger"}]

    want_main_area = 0
    want_control_area = 1

    def __init__(self):

         self.setFixedWidth(300)
         self.setFixedHeight(100)

         gui.separator(self.controlArea, height=20)
         gui.label(self.controlArea, self, "         LOST RAYS REMOVER", orientation="horizontal")
         gui.rubber(self.controlArea)

    def setBeam(self, beam):
        if ShadowCongruence.checkEmptyBeam(beam):
            output_beam = beam.duplicate(history=True)

            good = numpy.where(output_beam._beam.rays[:, 9] == 1)
            output_beam._beam.rays = copy.deepcopy(output_beam._beam.rays[good])

            self.send("Beam", output_beam)
            self.send("Trigger", TriggerIn(new_object=True))

if __name__ == "__main__":
    a = QtGui.QApplication(sys.argv)
    ow = BeamCleaner()
    ow.show()
    a.exec_()
    ow.saveSettings()



import sys, numpy

from PyQt5.QtGui import QPalette, QFont, QColor
from PyQt5.QtWidgets import QApplication, QMessageBox

from orangewidget.widget import OWAction
from orangewidget import gui
from orangewidget.settings import Setting

from oasys.widgets import widget
from oasys.widgets import gui as oasysgui
from oasys.widgets.gui import ConfirmDialog
from oasys.widgets import congruence

from oasys.util.oasys_util import TriggerIn, TriggerOut
from oasys.widgets.exchange import DataExchangeObject

import scipy.constants as codata

class EnergyBinning(object):
    def __init__(self,
                 energy_value_from = 0.0,
                 energy_value_to   = 0.0,
                 energy_value_central    = -1.0,
                 energy_value_half_power = -1.0,
                 energy_step = 0.0):
        self.energy_value_from       = energy_value_from
        self.energy_value_to         = energy_value_to
        self.energy_value_central    = energy_value_central
        self.energy_value_half_power = energy_value_half_power
        self.energy_step       = energy_step

    def __str__(self):
        return str(self.energy_value_from) + ", " + str(self.energy_value_to) + ", " + str(self.energy_step) + ", " + str(self.power_step)

class PowerLoopPoint(widget.OWWidget):

    name = "Power Density Loop Point"
    description = "Tools: LoopPoint"
    icon = "icons/cycle.png"
    maintainer = "Luca Rebuffi"
    maintainer_email = "lrebuffi(@at@)anl.gov"
    priority = 5
    category = "User Defined"
    keywords = ["data", "file", "load", "read"]

    inputs = [("Trigger", TriggerIn, "passTrigger"),
              ("ExchangeData", DataExchangeObject, "acceptExchangeData" )]

    outputs = [{"name":"Trigger",
                "type":TriggerOut,
                "doc":"Trigger",
                "id":"Trigger"}]
    want_main_area = 1

    current_new_object = 0
    number_of_new_objects = 0
    
    total_current_new_object = 0
    total_new_objects = Setting(0)

    run_loop = True
    suspend_loop = False

    energies = Setting("")

    seed_increment=Setting(1)

    autobinning = Setting(1)

    auto_n_step = Setting(1001)
    auto_perc_total_power = Setting(99)

    current_energy_binning = 0
    current_energy_value = None
    current_energy_value_central = None
    current_energy_value_half_power = None
    current_energy_step = None
    power_step = None

    energy_binnings = None

    test_mode = False

    external_binning = False

    #################################
    process_last = True
    #################################

    def __init__(self):
        self.runaction = OWAction("Start", self)
        self.runaction.triggered.connect(self.startLoop)
        self.addAction(self.runaction)

        self.runaction = OWAction("Stop", self)
        self.runaction.triggered.connect(self.stopLoop)
        self.addAction(self.runaction)

        self.runaction = OWAction("Suspend", self)
        self.runaction.triggered.connect(self.suspendLoop)
        self.addAction(self.runaction)

        self.runaction = OWAction("Restart", self)
        self.runaction.triggered.connect(self.restartLoop)
        self.addAction(self.runaction)

        self.setFixedWidth(1000)
        self.setFixedHeight(680)

        button_box = oasysgui.widgetBox(self.controlArea, "", addSpace=True, orientation="horizontal")

        self.start_button = gui.button(button_box, self, "Start", callback=self.startLoop)
        self.start_button.setFixedHeight(35)

        stop_button = gui.button(button_box, self, "Stop", callback=self.stopLoop)
        stop_button.setFixedHeight(35)
        font = QFont(stop_button.font())
        font.setBold(True)
        stop_button.setFont(font)
        palette = QPalette(stop_button.palette()) # make a copy of the palette
        palette.setColor(QPalette.ButtonText, QColor('red'))
        stop_button.setPalette(palette) # assign new palette

        self.stop_button = stop_button

        button_box = oasysgui.widgetBox(self.controlArea, "", addSpace=True, orientation="horizontal")

        suspend_button = gui.button(button_box, self, "Suspend", callback=self.suspendLoop)
        suspend_button.setFixedHeight(35)
        font = QFont(suspend_button.font())
        font.setBold(True)
        suspend_button.setFont(font)
        palette = QPalette(suspend_button.palette()) # make a copy of the palette
        palette.setColor(QPalette.ButtonText, QColor('orange'))
        suspend_button.setPalette(palette) # assign new palette

        self.re_start_button = gui.button(button_box, self, "Restart", callback=self.restartLoop)
        self.re_start_button.setFixedHeight(35)
        self.re_start_button.setEnabled(False)

        left_box_1 = oasysgui.widgetBox(self.controlArea, "Loop Management", addSpace=True, orientation="vertical", width=385, height=520)

        oasysgui.lineEdit(left_box_1, self, "seed_increment", "Source Montecarlo Seed Increment", labelWidth=250, valueType=int, orientation="horizontal")

        gui.separator(left_box_1)

        gui.comboBox(left_box_1, self, "autobinning", label="Energy Binning",
                                            items=["Manual", "Automatic"], labelWidth=260,
                                            callback=self.set_Autobinning, sendSelectedValue=False, orientation="horizontal")

        self.autobinning_box_1 = oasysgui.widgetBox(left_box_1, "", addSpace=False, orientation="vertical", height=80)
        self.autobinning_box_2 = oasysgui.widgetBox(left_box_1, "", addSpace=False, orientation="vertical", height=80)

        oasysgui.lineEdit(self.autobinning_box_1, self, "auto_n_step", "(Auto) Number of Steps", labelWidth=250, valueType=int, orientation="horizontal")
        oasysgui.lineEdit(self.autobinning_box_1, self, "auto_perc_total_power", "(Auto) % Total Power", labelWidth=250, valueType=float, orientation="horizontal")
        gui.button(self.autobinning_box_1, self, "Reload Spectrum", callback=self.read_spectrum_file)

        oasysgui.widgetLabel(self.autobinning_box_2, "\n\n\n\nEnergy From, Energy To, Energy Step [eV]")

        def write_text():
            self.energies = self.text_area.toPlainText()

        self.text_area = oasysgui.textArea(height=180, width=365, readOnly=False)
        self.text_area.setText(self.energies)
        self.text_area.setStyleSheet("background-color: white; font-family: Courier, monospace;")
        self.text_area.textChanged.connect(write_text)

        left_box_1.layout().addWidget(self.text_area)

        gui.separator(left_box_1)

        self.le_number_of_new_objects = oasysgui.lineEdit(left_box_1, self, "total_new_objects", "Total Energy Values", labelWidth=250, valueType=int, orientation="horizontal")
        self.le_number_of_new_objects.setReadOnly(True)
        font = QFont(self.le_number_of_new_objects.font())
        font.setBold(True)
        self.le_number_of_new_objects.setFont(font)
        palette = QPalette(self.le_number_of_new_objects.palette()) # make a copy of the palette
        palette.setColor(QPalette.Text, QColor('dark blue'))
        palette.setColor(QPalette.Base, QColor(243, 240, 160))
        self.le_number_of_new_objects.setPalette(palette)

        self.le_number_of_new_objects = oasysgui.lineEdit(left_box_1, self, "number_of_new_objects", "Current Binning Energy Values", labelWidth=250, valueType=int, orientation="horizontal")
        self.le_number_of_new_objects.setReadOnly(True)
        font = QFont(self.le_number_of_new_objects.font())
        font.setBold(True)
        self.le_number_of_new_objects.setFont(font)
        palette = QPalette(self.le_number_of_new_objects.palette()) # make a copy of the palette
        palette.setColor(QPalette.Text, QColor('dark blue'))
        palette.setColor(QPalette.Base, QColor(243, 240, 160))
        self.le_number_of_new_objects.setPalette(palette)

        gui.separator(left_box_1)

        le_current_value = oasysgui.lineEdit(left_box_1, self, "total_current_new_object", "Total New " + self.get_object_name(), labelWidth=250, valueType=int, orientation="horizontal")
        le_current_value.setReadOnly(True)
        font = QFont(le_current_value.font())
        font.setBold(True)
        le_current_value.setFont(font)
        palette = QPalette(le_current_value.palette()) # make a copy of the palette
        palette.setColor(QPalette.Text, QColor('dark blue'))
        palette.setColor(QPalette.Base, QColor(243, 240, 160))
        le_current_value.setPalette(palette)

        le_current_value = oasysgui.lineEdit(left_box_1, self, "current_new_object", "Current Binning New " + self.get_object_name(), labelWidth=250, valueType=int, orientation="horizontal")
        le_current_value.setReadOnly(True)
        font = QFont(le_current_value.font())
        font.setBold(True)
        le_current_value.setFont(font)
        palette = QPalette(le_current_value.palette()) # make a copy of the palette
        palette.setColor(QPalette.Text, QColor('dark blue'))
        palette.setColor(QPalette.Base, QColor(243, 240, 160))
        le_current_value.setPalette(palette)

        le_current_value = oasysgui.lineEdit(left_box_1, self, "current_energy_value", "Current Energy Value", labelWidth=250, valueType=float, orientation="horizontal")
        le_current_value.setReadOnly(True)
        font = QFont(le_current_value.font())
        font.setBold(True)
        le_current_value.setFont(font)
        palette = QPalette(le_current_value.palette()) # make a copy of the palette
        palette.setColor(QPalette.Text, QColor('dark blue'))
        palette.setColor(QPalette.Base, QColor(243, 240, 160))
        le_current_value.setPalette(palette)

        gui.rubber(self.controlArea)

        tabs = oasysgui.tabWidget(self.mainArea)
        tabs.setFixedHeight(self.height()-15)
        tabs.setFixedWidth(585)

        tab_plot = oasysgui.createTabPage(tabs, "Cumulated Power")

        self.cumulated_power_plot = oasysgui.plotWindow(tab_plot)
        self.cumulated_power_plot.setFixedHeight(self.height()-20)
        self.cumulated_power_plot.setFixedWidth(580)
        self.cumulated_power_plot.setGraphXLabel("Energy [eV]")
        self.cumulated_power_plot.setGraphYLabel("Cumulated Power [W]")
        self.cumulated_power_plot.setGraphTitle("Cumulated Power")

        self.set_Autobinning()

    def set_Autobinning(self):
        self.autobinning_box_1.setVisible(self.autobinning==1)
        self.autobinning_box_2.setVisible(self.autobinning==0)
        self.text_area.setReadOnly(self.autobinning==1)
        self.cumulated_power_plot.clear()
        self.cumulated_power_plot.setEnabled(self.autobinning==1)

    def read_spectrum_file(self):
        try:
            data = numpy.loadtxt("autobinning.dat", skiprows=1)

            calculated_data = DataExchangeObject(program_name="ShadowOui", widget_name="PowerLoopPoint")
            calculated_data.add_content("spectrum_data", data)

            self.acceptExchangeData(calculated_data)
        except Exception as e:
            QMessageBox.critical(self, "Error", str(e), QMessageBox.Ok)

            if self.IS_DEVELOP: raise e


    def acceptExchangeData(self, exchange_data):
        if not exchange_data is None:
            try:
                write_file = True

                try:
                    data = exchange_data.get_content("spectrum_data")
                    write_file = False
                except:
                    try:
                        data = exchange_data.get_content("srw_data")
                    except:
                        data = exchange_data.get_content("xoppy_data")

                energies                     = data[:, 0]
                flux_through_finite_aperture = data[:, 1]

                if write_file:
                    file = open("autobinning.dat", "w")
                    file.write("Energy Flux")

                    for energy, flux in zip(energies, flux_through_finite_aperture):
                        file.write("\n" + str(energy) + " " + str(flux))

                    file.flush()
                    file.close()

                if self.autobinning==0:
                    if write_file: QMessageBox.information(self, "Info", "File autobinning.dat written on working directory, switch to Automatic binning to load it", QMessageBox.Ok)
                else:
                    if write_file: QMessageBox.information(self, "Info", "File autobinning.dat written on working directory", QMessageBox.Ok)

                    congruence.checkStrictlyPositiveNumber(self.auto_n_step, "(Auto) % Number of Steps")
                    congruence.checkStrictlyPositiveNumber(self.auto_perc_total_power, "(Auto) % Total Power")

                    energy_step = energies[1]-energies[0]

                    power = flux_through_finite_aperture * (1e3 * energy_step * codata.e)
                    cumulated_power = numpy.cumsum(power)

                    total_power = cumulated_power[-1]

                    self.cumulated_power_plot.clear()
                    self.cumulated_power_plot.addCurve(energies, cumulated_power, replace=False, replot=True, legend="Cumulated Power")
                    self.cumulated_power_plot.setGraphXLabel("Energy [eV]")
                    self.cumulated_power_plot.setGraphYLabel("Cumulated Power [W]")
                    self.cumulated_power_plot.setGraphTitle("Total Power: " + str(round(total_power, 2)) + " W")

                    good = numpy.where(cumulated_power <= self.auto_perc_total_power*0.01*total_power)

                    energies        = energies[good]
                    cumulated_power = cumulated_power[good]

                    power_values = numpy.linspace(start=0, stop=numpy.max(cumulated_power), num=self.auto_n_step)
                    power_values = power_values[1:] # interpolating power=0 doesn't make sense

                    self.power_step = power_values[1]-power_values[0]

                    interpolated_upper_energies      = numpy.interp(power_values, cumulated_power, energies)
                    interpolated_half_power_energies = numpy.interp(power_values-0.5*self.power_step, cumulated_power, energies)
                    interpolated_lower_energies = numpy.insert(interpolated_upper_energies, 0, energies[0])
                    energy_bins = numpy.diff(interpolated_lower_energies)
                    interpolated_lower_energies = interpolated_lower_energies[:-1]
                    interpolated_central_energies = interpolated_upper_energies - (energy_bins/2)

                    self.energy_binnings = []
                    self.total_new_objects = 0

                    self.cumulated_power_plot.addCurve(interpolated_upper_energies, power_values, replace=False, replot=True, legend="Energy Binning",
                                                       color="red", linestyle=" ", symbol="+")

                    text = "Auto-Binning with Power Step: " + str(round(self.power_step, 3)) + \
                           "\nEnergy From, To, Central, Power/2, Bin\n------------------------------------------"

                    for energy_from, energy_to, energy_central, energy_half_power, energy_bin in zip(interpolated_lower_energies,
                                                                                                     interpolated_upper_energies,
                                                                                                     interpolated_central_energies,
                                                                                                     interpolated_half_power_energies,
                                                                                                     energy_bins):
                        energy_binning = EnergyBinning(energy_value_from=round(energy_from, 2),
                                                       energy_value_to=round(energy_to, 2),
                                                       energy_value_central=round(energy_central, 2),
                                                       energy_value_half_power=round(energy_half_power, 2),
                                                       energy_step=round(energy_bin, 3))

                        text += "\n" + \
                                str(round(energy_from, 2)) + ", " + \
                                str(round(energy_to, 2)) + ", " + \
                                str(round(energy_central, 2)) + ", " + \
                                str(round(energy_half_power, 2)) + ", " + \
                                str(round(energy_bin, 2))

                        self.energy_binnings.append(energy_binning)
                        self.total_new_objects += 1

                    self.text_area.setText(text)

                    self.external_binning = True
            except Exception as e:
                QMessageBox.critical(self, "Error", str(e), QMessageBox.Ok)
        else:
            self.energy_binnings = None
            self.total_new_objects = 0
            self.power_step = None
            self.external_binning = False
            self.text_area.setText("")

    def calculate_energy_binnings(self):
        if not self.external_binning:
            self.total_new_objects = 0
            self.power_step = None

            rows = self.energies.split("\n")
            for row in rows:
                data = row.split(",")
                if len(data) == 3:
                    if self.energy_binnings is None: self.energy_binnings = []

                    energy_binning = EnergyBinning(energy_value_from=float(data[0].strip()),
                                                   energy_value_to=float(data[1].strip()),
                                                   energy_step=float(data[2].strip()))
                    self.energy_binnings.append(energy_binning)
                    self.total_new_objects += int((energy_binning.energy_value_to - energy_binning.energy_value_from) / energy_binning.energy_step)

    def calculate_number_of_new_objects(self):
        if len(self.energy_binnings) > 0:
            if self.external_binning:
                self.number_of_new_objects = 1
            else:
                energy_binning = self.energy_binnings[self.current_energy_binning]

                self.number_of_new_objects = int((energy_binning.energy_value_to - energy_binning.energy_value_from) / energy_binning.energy_step)
        else:
            self.number_of_new_objects = 0

    def reset_values(self):
        self.current_new_object = 0
        self.total_current_new_object = 0
        self.current_energy_value = None
        self.current_energy_value_central = None
        self.current_energy_value_half_power = None
        self.current_energy_step = None
        self.current_energy_binning = 0
        self.current_power_step = None

        if not self.external_binning: self.energy_binnings = None

    def startLoop(self):
        try:
            self.calculate_energy_binnings()

            self.current_new_object = 1
            self.total_current_new_object = 1
            self.current_energy_binning = 0
            self.current_energy_value       = round(self.energy_binnings[0].energy_value_from, 8)
            self.current_energy_value_central    = round(self.energy_binnings[0].energy_value_central, 8)
            self.current_energy_value_half_power = round(self.energy_binnings[0].energy_value_half_power, 8)
            self.current_energy_step              = round(self.energy_binnings[0].energy_step, 8)

            self.calculate_number_of_new_objects()

            self.start_button.setEnabled(False)
            self.text_area.setEnabled(False)
            self.setStatusMessage("Running " + self.get_object_name() + " " + str(self.total_current_new_object) + " of " + str(self.total_new_objects))
            self.send("Trigger", TriggerOut(new_object=True,
                                            additional_parameters={"energy_value_from"       : self.current_energy_value,
                                                                   "energy_value_central"    : self.current_energy_value_central,
                                                                   "energy_value_half_power" : self.current_energy_value_half_power,
                                                                   "energy_step"             : self.current_energy_step,
                                                                   "power_step"              : -1 if self.power_step is None else self.power_step,
                                                                   "seed_increment"          : self.seed_increment,
                                                                   "test_mode" : False}))
        except Exception as e:
            if self.IS_DEVELOP : raise e
            else: pass

    def stopLoop(self):
        try:
            if ConfirmDialog.confirmed(parent=self, message="Confirm Interruption of the Loop?"):
                self.run_loop = False
                self.reset_values()
                self.setStatusMessage("Interrupted by user")
        except Exception as e:
            if self.IS_DEVELOP : raise e
            else: pass

    def suspendLoop(self):
        try:
            if ConfirmDialog.confirmed(parent=self, message="Confirm Suspension of the Loop?"):
                self.run_loop = False
                self.suspend_loop = True
                self.stop_button.setEnabled(False)
                self.re_start_button.setEnabled(True)
                self.setStatusMessage("Suspended by user")
        except Exception as e:
            if self.IS_DEVELOP : raise e
            else: pass


    def restartLoop(self):
        try:
            self.run_loop = True
            self.suspend_loop = False
            self.stop_button.setEnabled(True)
            self.re_start_button.setEnabled(False)
            self.passTrigger(TriggerIn(new_object=True))
        except Exception as e:
            if self.IS_DEVELOP : raise e
            else: pass

    def get_object_name(self):
        return "Beam"

    def passTrigger(self, trigger):
        if self.run_loop:
            if trigger:
                if trigger.interrupt:
                    self.reset_values()
                    self.start_button.setEnabled(True)
                    self.text_area.setEnabled(True)
                    self.setStatusMessage("")
                    self.send("Trigger", TriggerOut(new_object=False))
                elif trigger.new_object:
                    if self.energy_binnings is None: self.calculate_energy_binnings()

                    if self.current_energy_binning < len(self.energy_binnings):
                        energy_binning = self.energy_binnings[self.current_energy_binning]

                        self.total_current_new_object += 1

                        if self.current_new_object < self.number_of_new_objects:
                            if self.current_energy_value is None:
                                self.current_new_object = 1
                                self.calculate_number_of_new_objects()
                                self.current_energy_value            = round(energy_binning.energy_value_from, 8)
                                self.current_energy_value_central    = round(energy_binning.energy_value_central, 8)
                                self.current_energy_value_half_power = round(energy_binning.energy_value_half_power, 8)
                            else:
                                self.current_new_object += 1
                                self.current_energy_value = round(self.current_energy_value + energy_binning.energy_step, 8)

                            self.setStatusMessage("Running " + self.get_object_name() + " " + str(self.total_current_new_object) + " of " + str(self.total_new_objects))
                            self.start_button.setEnabled(False)
                            self.text_area.setEnabled(False)
                            self.send("Trigger", TriggerOut(new_object=True,
                                                            additional_parameters={"energy_value"            : self.current_energy_value,
                                                                                   "energy_value_central"    : self.current_energy_value_central,
                                                                                   "energy_value_half_power" : self.current_energy_value_half_power,
                                                                                   "energy_step" : energy_binning.energy_step,
                                                                                   "power_step" : -1 if self.power_step is None else self.power_step,
                                                                                   "seed_increment" : self.seed_increment,
                                                                                   "test_mode" : False}))
                        else:
                            self.current_energy_binning += 1

                            if self.current_energy_binning < len(self.energy_binnings):
                                energy_binning = self.energy_binnings[self.current_energy_binning]

                                self.current_new_object = 1
                                self.calculate_number_of_new_objects()
                                self.current_energy_value = round(energy_binning.energy_value_from, 8)
                                self.current_energy_value_central    = round(energy_binning.energy_value_central, 8)
                                self.current_energy_value_half_power = round(energy_binning.energy_value_half_power, 8)

                                self.setStatusMessage("Running " + self.get_object_name() + " " + str(self.total_current_new_object) + " of " + str(self.total_new_objects))
                                self.start_button.setEnabled(False)
                                self.text_area.setEnabled(False)
                                self.send("Trigger", TriggerOut(new_object=True,
                                                                additional_parameters={"energy_value" : self.current_energy_value,
                                                                                       "energy_value_central"    : self.current_energy_value_central,
                                                                                       "energy_value_half_power" : self.current_energy_value_half_power,
                                                                                       "energy_step" : energy_binning.energy_step,
                                                                                       "power_step" : -1 if self.power_step is None else self.power_step,
                                                                                       "seed_increment" : self.seed_increment,
                                                                                       "test_mode" : False}))
                            else:
                                self.reset_values()
                                self.start_button.setEnabled(True)
                                self.text_area.setEnabled(True)
                                self.setStatusMessage("")
                                self.send("Trigger", TriggerOut(new_object=False))
                    else:
                        self.reset_values()
                        self.start_button.setEnabled(True)
                        self.text_area.setEnabled(True)
                        self.setStatusMessage("")
                        self.send("Trigger", TriggerOut(new_object=False))
        else:
            if not self.suspend_loop:
                self.reset_values()
                self.start_button.setEnabled(True)
                self.text_area.setEnabled(True)

            self.send("Trigger", TriggerOut(new_object=False))
            self.setStatusMessage("")
            self.suspend_loop = False
            self.run_loop = True

if __name__ == "__main__":
    a = QApplication(sys.argv)
    ow = PowerLoopPoint()
    ow.show()
    a.exec_()
    ow.saveSettings()



## @package random_distributions_demo
## Demonstrates to generation of numbers from an arbitrary distribution.
##

import numpy
from orangecontrib.aps.util.random_distributions import *
from orangecontrib.aps.util.enhanced_grid import Grid2D
from random import random

from matplotlib import pyplot as plt

def index_of(array, value):
    array = numpy.asarray(array)
    idx = (numpy.abs(array - value)).argmin()
    return idx

# Demonstrates to generation of points from a 2D distribution.
# An image is produced that shows an estimate of the distribution
# for a samnple of points generate from the specified distribution.
def demo_distribution_2d():
    data = numpy.loadtxt("/Users/lrebuffi/TEST.dat", skiprows=1)

    x_values = numpy.unique(data[:, 0])
    y_values = numpy.unique(data[:, 1])

    min_x = numpy.min(x_values)
    max_x = numpy.max(x_values)
    delta_x =  numpy.max(x_values) -  numpy.min(x_values)

    min_y = numpy.min(y_values)
    max_y = numpy.max(y_values)
    delta_y =  numpy.max(y_values) -  numpy.min(y_values)

    dim_x = len(x_values)
    dim_y = len(y_values)

    z_values = numpy.zeros((dim_x, dim_y))

    for i in range(dim_x):
        for j in range(dim_y):
            index = i*dim_x + j

            z_values[i, j] = data[index, 2]

    fig, ax = plt.subplots(ncols=2)
    ax[0].imshow(z_values, interpolation='bilinear', origin='lower', extent=[0, 100, 0, 100],
              vmax=numpy.max(z_values), vmin=numpy.min(z_values))

    grid = Grid2D((dim_x, dim_y))
    grid[..., ...] = z_values.tolist()

    probabilities = distribution_from_grid(grid, dim_x, dim_y)

    d = Distribution2D(probabilities, (min_x, min_y), (max_x, max_y))

    samples = []

    for k in range(1000000):
        samples.append(d(random(), random()))

    z_values_rand = numpy.zeros((dim_x, dim_y))

    for sample in samples:
        x_value = sample[0]
        y_value = sample[1]

        i = index_of(x_values, min_x + x_value*delta_x)
        j = index_of(y_values, min_y + y_value*delta_y)
        
        z_values_rand[i, j] += 1

    ax[1].imshow(z_values_rand, interpolation='bilinear', origin='lower', extent=[0, 100, 0, 100],
              vmax=numpy.max(z_values_rand), vmin=numpy.min(z_values_rand))
    plt.show()



#Run any of these to see how they work.
#demo_distribution_1d()
demo_distribution_2d()



import numpy
from numpy.matlib import repmat
from scipy.signal import convolve2d

from PyQt5 import QtWidgets
from PyQt5.QtGui import QPalette, QColor, QFont
from orangewidget import gui
from orangewidget.settings import Setting
from oasys.widgets import gui as oasysgui
from oasys.widgets import congruence
from orangewidget import widget
from oasys.util.oasys_util import TriggerOut, EmittingStream

from syned.beamline.beamline import Beamline
from syned.beamline.optical_elements.absorbers.slit import Slit
from syned.storage_ring.light_source import ElectronBeam, LightSource
from syned.widget.widget_decorator import WidgetDecorator
from syned.beamline.shape import Rectangle

from orangecontrib.shadow.util.shadow_objects import ShadowBeam, ShadowSource
from orangecontrib.shadow.widgets.gui.ow_generic_element import GenericElement

from orangecontrib.aps.util.random_distributions import Distribution2D, Grid2D, distribution_from_grid
from orangecontrib.aps.util.custom_distribution import CustomDistribution

import scipy.constants as codata

m2ev = codata.c * codata.h / codata.e

from srwlib import *

class Distribution:
    POSITION = 0
    DIVERGENCE = 1

class APSUndulator(GenericElement):

    TABS_AREA_HEIGHT = 620

    name = "APS Undulator"
    description = "Shadow Source: APS Undulator"
    icon = "icons/undulator.png"
    priority = 1
    maintainer = "Luca Rebuffi"
    maintainer_email = "luca.rebuffi(@at@)elettra.eu"
    category = "Sources"
    keywords = ["data", "file", "load", "read"]

    inputs = WidgetDecorator.syned_input_data()
    inputs.append(("SynedData#2", Beamline, "receive_syned_data"))
    inputs.append(("Trigger", TriggerOut, "sendNewBeam"))

    outputs = [{"name":"Beam",
                "type":ShadowBeam,
                "doc":"Shadow Beam",
                "id":"beam"}]

    distribution_source = Setting(0)

    # SRW INPUT

    number_of_periods = Setting(184) # Number of ID Periods (without counting for terminations
    undulator_period =  Setting(0.025) # Period Length [m]
    Kv =  Setting(0.857)
    Kh =  Setting(0)

    electron_energy_in_GeV = Setting(6.0)
    electron_energy_spread = Setting(1.35e-3)
    ring_current = Setting(0.2)
    electron_beam_size_h = Setting(1.45e-05)
    electron_beam_size_v = Setting(2.8e-06)
    electron_beam_divergence_h = Setting(2.9e-06)
    electron_beam_divergence_v = Setting(1.5e-06)

    source_dimension_wf_h_slit_gap = Setting(0.0015)
    source_dimension_wf_v_slit_gap = Setting(0.0015)
    source_dimension_wf_h_slit_points=Setting(301)
    source_dimension_wf_v_slit_points=Setting(301)
    source_dimension_wf_distance = Setting(28.0)

    horizontal_range_modification_factor_at_resizing       = Setting(0.5)
    horizontal_resolution_modification_factor_at_resizing  = Setting(5.0)
    vertical_range_modification_factor_at_resizing         = Setting(0.5)
    vertical_resolution_modification_factor_at_resizing    = Setting(5.0)

    kind_of_sampler = Setting(1)
    save_srw_result = Setting(1)
    
    # SRW FILE INPUT

    source_dimension_srw_file     = Setting("intensity_source_dimension.dat")
    angular_distribution_srw_file = Setting("intensity_angular_distribution.dat")

    # ASCII FILE INPUT

    x_positions_file = Setting("x_positions.txt")
    z_positions_file = Setting("z_positions.txt")
    x_positions_factor = Setting(0.01)
    z_positions_factor = Setting(0.01)
    x_divergences_file = Setting("x_divergences.txt")
    z_divergences_file = Setting("z_divergences.txt")
    x_divergences_factor = Setting(1.0)
    z_divergences_factor = Setting(1.0)

    combine_strategy = Setting(0)

    # SHADOW SETTINGS

    number_of_rays=Setting(5000)
    seed=Setting(6775431)

    use_harmonic = Setting(0)
    harmonic_number = Setting(1)
    harmonic_energy = 0.0
    energy=Setting(10000.0)

    polarization = Setting(1)
    coherent_beam = Setting(0)
    phase_diff = Setting(0.0)
    polarization_degree = Setting(1.0)

    optimize_source=Setting(0)
    optimize_file_name = Setting("NONESPECIFIED")
    max_number_of_rejected_rays = Setting(10000000)

    file_to_write_out = Setting(0)

    auto_energy = Setting(0.0)
    auto_harmonic_number = Setting(1)

    energy_step = None
    power_step = None
    compute_power = False
    test_mode = False

    def __init__(self, show_automatic_box=False):
        super().__init__(show_automatic_box=show_automatic_box)

        self.runaction = widget.OWAction("Run Shadow/Source", self)
        self.runaction.triggered.connect(self.runShadowSource)
        self.addAction(self.runaction)

        self.general_options_box.setVisible(False)

        button_box = oasysgui.widgetBox(self.controlArea, "", addSpace=False, orientation="horizontal")

        button = gui.button(button_box, self, "Run Shadow/Source", callback=self.runShadowSource)
        font = QFont(button.font())
        font.setBold(True)
        button.setFont(font)
        palette = QPalette(button.palette()) # make a copy of the palette
        palette.setColor(QPalette.ButtonText, QColor('Dark Blue'))
        button.setPalette(palette) # assign new palette
        button.setFixedHeight(45)

        button = gui.button(button_box, self, "Reset Fields", callback=self.callResetSettings)
        font = QFont(button.font())
        font.setItalic(True)
        button.setFont(font)
        palette = QPalette(button.palette()) # make a copy of the palette
        palette.setColor(QPalette.ButtonText, QColor('Dark Red'))
        button.setPalette(palette) # assign new palette
        button.setFixedHeight(45)
        button.setFixedWidth(150)

        gui.separator(self.controlArea)

        ######################################

        self.controlArea.setFixedWidth(self.CONTROL_AREA_WIDTH)

        tabs_setting = oasysgui.tabWidget(self.controlArea)
        tabs_setting.setFixedHeight(self.TABS_AREA_HEIGHT)
        tabs_setting.setFixedWidth(self.CONTROL_AREA_WIDTH-5)

        tab_shadow = oasysgui.createTabPage(tabs_setting, "Shadow Setting")
        tab_spdiv = oasysgui.createTabPage(tabs_setting, "Position/Divergence Setting")
        tab_util = oasysgui.createTabPage(tabs_setting, "Utility")

        gui.comboBox(tab_spdiv, self, "distribution_source", label="Distribution Source", labelWidth=310,
                     items=["SRW Calculation", "SRW Files", "ASCII Files"], orientation="horizontal", callback=self.set_DistributionSource)

        self.srw_box = oasysgui.widgetBox(tab_spdiv, "", addSpace=False, orientation="vertical", height=550)
        self.srw_files_box = oasysgui.widgetBox(tab_spdiv, "", addSpace=False, orientation="vertical", height=550)
        self.ascii_box = oasysgui.widgetBox(tab_spdiv, "", addSpace=False, orientation="vertical", height=550)

        self.set_DistributionSource()

        ####################################################################################
        # SHADOW

        left_box_1 = oasysgui.widgetBox(tab_shadow, "Monte Carlo and Energy Spectrum", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(left_box_1, self, "number_of_rays", "Number of Rays", tooltip="Number of Rays", labelWidth=250, valueType=int, orientation="horizontal")
        oasysgui.lineEdit(left_box_1, self, "seed", "Seed", tooltip="Seed (0=clock)", labelWidth=250, valueType=int, orientation="horizontal")

        gui.comboBox(left_box_1, self, "use_harmonic", label="Photon Energy Setting",
                     items=["Harmonic", "Other"], labelWidth=260,
                     callback=self.set_WFUseHarmonic, sendSelectedValue=False, orientation="horizontal")

        self.use_harmonic_box_1 = oasysgui.widgetBox(left_box_1, "", addSpace=False, orientation="vertical", height=50)
        oasysgui.lineEdit(self.use_harmonic_box_1, self, "harmonic_number", "Harmonic #", labelWidth=260, valueType=int, orientation="horizontal", callback=self.set_harmonic_energy)
        le_he = oasysgui.lineEdit(self.use_harmonic_box_1, self, "harmonic_energy", "Harmonic Energy", labelWidth=260, valueType=float, orientation="horizontal")
        le_he.setReadOnly(True)
        font = QFont(le_he.font())
        font.setBold(True)
        le_he.setFont(font)
        palette = QPalette(le_he.palette())
        palette.setColor(QPalette.Text, QColor('dark blue'))
        palette.setColor(QPalette.Base, QColor(243, 240, 160))
        le_he.setPalette(palette)

        self.use_harmonic_box_2 = oasysgui.widgetBox(left_box_1, "", addSpace=False, orientation="vertical", height=50)
        oasysgui.lineEdit(self.use_harmonic_box_2, self, "energy", "Photon Energy [eV]", labelWidth=260, valueType=float, orientation="horizontal")

        self.set_WFUseHarmonic()

        polarization_box = oasysgui.widgetBox(tab_shadow, "Polarization", addSpace=False, orientation="vertical", height=140)

        gui.comboBox(polarization_box, self, "polarization", label="Polarization", labelWidth=310,
                     items=["No", "Yes"], orientation="horizontal", callback=self.set_Polarization)

        self.ewp_box_8 = oasysgui.widgetBox(polarization_box, "", addSpace=False, orientation="vertical")

        gui.comboBox(self.ewp_box_8, self, "coherent_beam", label="Coherent Beam", labelWidth=310,
                     items=["No", "Yes"], orientation="horizontal")

        oasysgui.lineEdit(self.ewp_box_8, self, "phase_diff", "Phase Difference [deg,0=linear,+90=ell/right]", labelWidth=310, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.ewp_box_8, self, "polarization_degree", "Polarization Degree [cos_s/(cos_s+sin_s)]", labelWidth=310, valueType=float, orientation="horizontal")

        self.set_Polarization()

        ##############################

        left_box_4 = oasysgui.widgetBox(tab_shadow, "Reject Rays", addSpace=False, orientation="vertical", height=140)

        gui.comboBox(left_box_4, self, "optimize_source", label="Optimize Source", items=["No", "Using file with phase/space volume)", "Using file with slit/acceptance"],
                     labelWidth=120, callback=self.set_OptimizeSource, orientation="horizontal")
        self.optimize_file_name_box       = oasysgui.widgetBox(left_box_4, "", addSpace=False, orientation="vertical", height=80)

        file_box = oasysgui.widgetBox(self.optimize_file_name_box, "", addSpace=True, orientation="horizontal", height=25)

        self.le_optimize_file_name = oasysgui.lineEdit(file_box, self, "optimize_file_name", "File Name", labelWidth=100,  valueType=str, orientation="horizontal")

        gui.button(file_box, self, "...", callback=self.selectOptimizeFile)

        oasysgui.lineEdit(self.optimize_file_name_box, self, "max_number_of_rejected_rays", "Max number of rejected rays (set 0 for infinity)", labelWidth=280,  valueType=int, orientation="horizontal")

        self.set_OptimizeSource()

        adv_other_box = oasysgui.widgetBox(tab_shadow, "Optional file output", addSpace=False, orientation="vertical")

        gui.comboBox(adv_other_box, self, "file_to_write_out", label="Files to write out", labelWidth=120,
                     items=["None", "Begin.dat", "Debug (begin.dat + start.xx/end.xx)"],
                     sendSelectedValue=False, orientation="horizontal")

        ####################################################################################
        # SRW

        tabs_srw = oasysgui.tabWidget(self.srw_box)

        gui.comboBox(self.srw_box, self, "kind_of_sampler", label="Random Generator", labelWidth=250,
                     items=["Simple (Fast)", "Accurate (Slow)"], orientation="horizontal")

        gui.comboBox(self.srw_box, self, "save_srw_result", label="Save SRW results", labelWidth=310,
                     items=["No", "Yes"], orientation="horizontal", callback=self.set_SaveFileSRW)

        self.save_file_box = oasysgui.widgetBox(self.srw_box, "", addSpace=False, orientation="vertical")
        self.save_file_box_empty = oasysgui.widgetBox(self.srw_box, "", addSpace=False, orientation="vertical", height=55)

        file_box = oasysgui.widgetBox(self.save_file_box, "", addSpace=False, orientation="horizontal", height=25)

        self.le_source_dimension_srw_file = oasysgui.lineEdit(file_box, self, "source_dimension_srw_file", "Source Dimension File", labelWidth=140,  valueType=str, orientation="horizontal")

        gui.button(file_box, self, "...", callback=self.selectSourceDimensionFile)

        file_box = oasysgui.widgetBox(self.save_file_box, "", addSpace=False, orientation="horizontal", height=25)

        self.le_angular_distribution_srw_file = oasysgui.lineEdit(file_box, self, "angular_distribution_srw_file", "Angular Distribution File", labelWidth=140,  valueType=str, orientation="horizontal")

        gui.button(file_box, self, "...", callback=self.selectAngularDistributionFile)

        self.set_SaveFileSRW()

        tab_ls = oasysgui.createTabPage(tabs_srw, "Undulator Setting")
        tab_wf = oasysgui.createTabPage(tabs_srw, "Wavefront Setting")

        left_box_1 = oasysgui.widgetBox(tab_ls, "Undulator Parameters", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(left_box_1, self, "number_of_periods", "Number of Periods", labelWidth=260,  valueType=float, orientation="horizontal")
        oasysgui.lineEdit(left_box_1, self, "undulator_period", "Undulator Period [m]", labelWidth=260,  valueType=float, orientation="horizontal", callback=self.set_harmonic_energy)
        oasysgui.lineEdit(left_box_1, self, "Kv", "K Vertical", labelWidth=260,  valueType=float, orientation="horizontal", callback=self.set_harmonic_energy)
        oasysgui.lineEdit(left_box_1, self, "Kh", "K Horizontal", labelWidth=260,  valueType=float, orientation="horizontal", callback=self.set_harmonic_energy)

        left_box_2 = oasysgui.widgetBox(tab_ls, "Machine Parameters", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(left_box_2, self, "electron_energy_in_GeV", "Energy [GeV]", labelWidth=260, valueType=float, orientation="horizontal", callback=self.set_harmonic_energy)
        oasysgui.lineEdit(left_box_2, self, "electron_energy_spread", "Energy Spread", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(left_box_2, self, "ring_current", "Ring Current [A]", labelWidth=260, valueType=float, orientation="horizontal")
        
        #gui.separator(left_box_2)
        
        oasysgui.lineEdit(left_box_2, self, "electron_beam_size_h",       "Horizontal Beam Size [m]", labelWidth=230, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(left_box_2, self, "electron_beam_size_v",       "Vertical Beam Size [m]",  labelWidth=230, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(left_box_2, self, "electron_beam_divergence_h", "Horizontal Beam Divergence [rad]", labelWidth=230, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(left_box_2, self, "electron_beam_divergence_v", "Vertical Beam Divergence [rad]", labelWidth=230, valueType=float, orientation="horizontal")

        left_box_3 = oasysgui.widgetBox(tab_wf, "Wavefront Propagation Parameters", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(left_box_3, self, "source_dimension_wf_h_slit_gap", "H Slit Gap [m]", labelWidth=250, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(left_box_3, self, "source_dimension_wf_v_slit_gap", "V Slit Gap [m]", labelWidth=250, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(left_box_3, self, "source_dimension_wf_h_slit_points", "H Slit Points", labelWidth=250, valueType=int, orientation="horizontal")
        oasysgui.lineEdit(left_box_3, self, "source_dimension_wf_v_slit_points", "V Slit Points", labelWidth=250, valueType=int, orientation="horizontal")
        oasysgui.lineEdit(left_box_3, self, "source_dimension_wf_distance", "Propagation Distance [m]", labelWidth=250, valueType=float, orientation="horizontal")

        left_box_4 = oasysgui.widgetBox(tab_wf, "Drift Back Propagation Parameters", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(left_box_4, self, "horizontal_range_modification_factor_at_resizing", "H range modification factor at resizing", labelWidth=290, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(left_box_4, self, "horizontal_resolution_modification_factor_at_resizing", "H resolution modification factor at resizing", labelWidth=290, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(left_box_4, self, "vertical_range_modification_factor_at_resizing", "V range modification factor at resizing", labelWidth=290, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(left_box_4, self, "vertical_resolution_modification_factor_at_resizing", "V resolution modification factor at resizing", labelWidth=290, valueType=float, orientation="horizontal")


        ####################################################################################
        # SRW FILES

        gui.separator(self.srw_files_box)

        file_box = oasysgui.widgetBox(self.srw_files_box, "", addSpace=True, orientation="horizontal", height=45)

        self.le_source_dimension_srw_file = oasysgui.lineEdit(file_box, self, "source_dimension_srw_file", "Source Dimension File", labelWidth=180,  valueType=str, orientation="vertical")

        gui.button(file_box, self, "...", height=45, callback=self.selectSourceDimensionFile)

        file_box = oasysgui.widgetBox(self.srw_files_box, "", addSpace=True, orientation="horizontal", height=45)

        self.le_angular_distribution_srw_file = oasysgui.lineEdit(file_box, self, "angular_distribution_srw_file", "Angular Distribution File", labelWidth=180,  valueType=str, orientation="vertical")

        gui.button(file_box, self, "...", height=45, callback=self.selectAngularDistributionFile)


        ####################################################################################
        # ASCII FILES

        gui.separator(self.ascii_box)

        file_box = oasysgui.widgetBox(self.ascii_box, "", addSpace=True, orientation="horizontal", height=45)

        self.le_x_positions_file = oasysgui.lineEdit(file_box, self, "x_positions_file", "X Positions File", labelWidth=180,  valueType=str, orientation="vertical")

        gui.button(file_box, self, "...", height=45, callback=self.selectXPositionsFile)

        file_box = oasysgui.widgetBox(self.ascii_box, "", addSpace=True, orientation="horizontal", height=45)

        self.le_z_positions_file = oasysgui.lineEdit(file_box, self, "z_positions_file", "Z Positions File", labelWidth=180,  valueType=str, orientation="vertical")

        gui.button(file_box, self, "...", height=45, callback=self.selectZPositionsFile)

        file_box = oasysgui.widgetBox(self.ascii_box, "", addSpace=True, orientation="horizontal", height=45)

        self.le_x_divergences_file = oasysgui.lineEdit(file_box, self, "x_divergences_file", "X Divergences File", labelWidth=180,  valueType=str, orientation="vertical")

        gui.button(file_box, self, "...", height=45, callback=self.selectXDivergencesFile)

        file_box = oasysgui.widgetBox(self.ascii_box, "", addSpace=True, orientation="horizontal", height=45)

        self.le_z_divergences_file = oasysgui.lineEdit(file_box, self, "z_divergences_file", "Z Divergences File", labelWidth=180,  valueType=str, orientation="vertical")

        gui.button(file_box, self, "...", height=45, callback=self.selectZDivergencesFile)

        gui.separator(self.ascii_box)

        oasysgui.lineEdit(self.ascii_box, self, "x_positions_factor",   "X Positions UM to Workspace UM", labelWidth=230, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.ascii_box, self, "z_positions_factor",   "Z Positions UM to Workspace UM",  labelWidth=230, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.ascii_box, self, "x_divergences_factor", "X Divergences UM to rad", labelWidth=230, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.ascii_box, self, "z_divergences_factor", "X Divergences UM to rad", labelWidth=230, valueType=float, orientation="horizontal")

        gui.separator(self.ascii_box)

        gui.comboBox(self.ascii_box, self, "combine_strategy", label="2D Distribution Creation Strategy", labelWidth=310,
                     items=["Sqrt(Product)", "Sqrt(Quadratic Sum)", "Convolution", "Average"], orientation="horizontal", callback=self.set_SaveFileSRW)

        ####################################################################################
        # Utility

        left_box_1 = oasysgui.widgetBox(tab_util, "Auto Setting of Undulator", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(left_box_1, self, "auto_energy", "Set Undulator at Energy [eV]", labelWidth=250, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(left_box_1, self, "auto_harmonic_number", "As Harmonic #",  labelWidth=250, valueType=int, orientation="horizontal")

        gui.button(left_box_1, self, "Set Kv value", callback=self.auto_set_undulator)

        gui.rubber(self.controlArea)
        gui.rubber(self.mainArea)



    ####################################################################################
    # GRAPHICS
    ####################################################################################

    def after_change_workspace_units(self):
        pass

    def auto_set_undulator(self):
        if not self.distribution_source == 0: raise Exception("This calculation can be performed only for explicit SRW Calculation")
        congruence.checkStrictlyPositiveNumber(self.auto_energy, "Set Undulator at Energy")
        congruence.checkStrictlyPositiveNumber(self.auto_harmonic_number, "As Harmonic #")
        congruence.checkStrictlyPositiveNumber(self.electron_energy_in_GeV, "Energy")
        congruence.checkStrictlyPositiveNumber(self.undulator_period, "Period Length")

        wavelength = self.auto_harmonic_number*m2ev/self.auto_energy

        self.Kv = round(numpy.sqrt(2*(((wavelength*2*self.gamma()**2)/self.undulator_period)-1)), 6)
        self.Kh = 0

        self.set_WFUseHarmonic()

    def set_harmonic_energy(self):
        if self.distribution_source==0 and self.use_harmonic==0:
            self.harmonic_energy = round(self.resonance_energy(harmonic=self.harmonic_number), 2)
        else:
            self.harmonic_energy = numpy.nan

    def get_write_file_options(self):
        write_begin_file = 0
        write_start_file = 0
        write_end_file = 0

        if self.file_to_write_out == 1:
            write_begin_file = 1
        if self.file_to_write_out == 2:
            write_begin_file = 1

            if sys.platform == 'linux':
                QtWidgets.QMessageBox.warning(self, "Warning", "Debug Mode is not yet available for sources in Linux platforms", QtWidgets.QMessageBox.Ok)
            else:
                write_start_file = 1
                write_end_file = 1

        return write_begin_file, write_start_file, write_end_file

    def set_WFUseHarmonic(self):
        self.use_harmonic_box_1.setVisible(self.use_harmonic==0)
        self.use_harmonic_box_2.setVisible(self.use_harmonic==1)

        self.set_harmonic_energy()

    def set_DistributionSource(self):
        self.srw_box.setVisible(self.distribution_source == 0)
        self.srw_files_box.setVisible(self.distribution_source == 1)
        self.ascii_box.setVisible(self.distribution_source == 2)

        self.set_harmonic_energy()

    def set_Polarization(self):
        self.ewp_box_8.setVisible(self.polarization==1)

    def set_OptimizeSource(self):
        self.optimize_file_name_box.setVisible(self.optimize_source != 0)

    def set_SaveFileSRW(self):
        self.save_file_box.setVisible(self.save_srw_result == 1)
        self.save_file_box_empty.setVisible(self.save_srw_result == 0)

    def selectOptimizeFile(self):
        self.le_optimize_file_name.setText(oasysgui.selectFileFromDialog(self, self.optimize_file_name, "Open Optimize Source Parameters File"))

    def selectSourceDimensionFile(self):
        self.le_source_dimension_srw_file.setText(oasysgui.selectFileFromDialog(self, self.source_dimension_srw_file, "Open Source Dimension File"))

    def selectAngularDistributionFile(self):
        self.le_angular_distribution_srw_file.setText(oasysgui.selectFileFromDialog(self, self.angular_distribution_srw_file, "Open Angular Distribution File"))

    def selectXPositionsFile(self):
        self.le_x_positions_file.setText(oasysgui.selectFileFromDialog(self, self.x_positions_file, "Open X Positions File", file_extension_filter="*.dat, *.txt"))

    def selectZPositionsFile(self):
        self.le_z_positions_file.setText(oasysgui.selectFileFromDialog(self, self.z_positions_file, "Open Z Positions File", file_extension_filter="*.dat, *.txt"))

    def selectXDivergencesFile(self):
        self.le_x_divergences_file.setText(oasysgui.selectFileFromDialog(self, self.x_divergences_file, "Open X Divergences File", file_extension_filter="*.dat, *.txt"))

    def selectZDivergencesFile(self):
        self.le_z_divergences_file.setText(oasysgui.selectFileFromDialog(self, self.z_divergences_file, "Open Z Divergences File", file_extension_filter="*.dat, *.txt"))

    ####################################################################################
    # SYNED
    ####################################################################################

    def receive_syned_data(self, data):
        if not data is None:
            try:
                if data.get_beamline_elements_number() > 0:
                    slit_element = data.get_beamline_element_at(0)
                    slit = slit_element.get_optical_element()
                    coordinates = slit_element.get_coordinates()

                    if isinstance(slit, Slit) and isinstance(slit.get_boundary_shape(), Rectangle):
                        rectangle = slit.get_boundary_shape()

                        self.source_dimension_wf_h_slit_gap = rectangle._x_right - rectangle._x_left
                        self.source_dimension_wf_v_slit_gap = rectangle._y_top - rectangle._y_bottom
                        self.source_dimension_wf_distance = coordinates.p()
                elif not data._light_source is None and isinstance(data._light_source, LightSource):
                    light_source = data._light_source

                    self.electron_energy_in_GeV = light_source._electron_beam._energy_in_GeV
                    self.electron_energy_spread = light_source._electron_beam._energy_spread
                    self.ring_current = light_source._electron_beam._current

                    x, xp, y, yp = light_source._electron_beam.get_sigmas_all()

                    self.electron_beam_size_h = round(x, 9)
                    self.electron_beam_size_v = round(y, 9)
                    self.electron_beam_divergence_h = round(xp, 10)
                    self.electron_beam_divergence_v = round(yp, 10)

                    self.Kh = light_source._magnetic_structure._K_horizontal
                    self.Kv = light_source._magnetic_structure._K_vertical
                    self.undulator_period = light_source._magnetic_structure._period_length
                    self.number_of_periods = light_source._magnetic_structure._number_of_periods
                else:
                    raise ValueError("Syned data not correct")
            except Exception as exception:
                QtWidgets.QMessageBox.critical(self, "Error", str(exception), QtWidgets.QMessageBox.Ok)

    def receive_specific_syned_data(self, data):
        raise NotImplementedError()



    ####################################################################################
    # PROCEDURES
    ####################################################################################

    def runShadowSource(self):
        self.setStatusMessage("")
        self.progressBarInit()

        sys.stdout = EmittingStream(textWritten=self.writeStdOut)

        try:
            if self.test_mode:
                beam_out = ShadowBeam()
                beam_out._beam.rays = numpy.zeros((1000, 18))
                total_power = 100.0
            else:
                self.checkFields()

                ###########################################
                # TODO: TO BE ADDED JUST IN CASE OF BROKEN
                #       ENVIRONMENT: MUST BE FOUND A PROPER WAY
                #       TO TEST SHADOW
                self.fixWeirdShadowBug()
                ###########################################

                shadow_src = ShadowSource.create_src()

                self.populateFields(shadow_src)

                self.progressBarSet(10)

                self.setStatusMessage("Running SHADOW")

                write_begin_file, write_start_file, write_end_file = self.get_write_file_options()

                beam_out = ShadowBeam.traceFromSource(shadow_src,
                                                      write_begin_file=write_begin_file,
                                                      write_start_file=write_start_file,
                                                      write_end_file=write_end_file)

                self.fix_Intensity(beam_out)

                self.progressBarSet(20)

                if self.distribution_source == 0:
                    self.setStatusMessage("Running SRW")

                    x, z, intensity_source_dimension, x_first, z_first, intensity_angular_distribution, total_power = self.runSRWCalculation()
                elif self.distribution_source == 1:
                    self.setStatusMessage("Loading SRW files")

                    x, z, intensity_source_dimension, x_first, z_first, intensity_angular_distribution = self.loadSRWFiles()
                elif self.distribution_source == 2: # ASCII FILES
                    self.setStatusMessage("Loading Ascii files")

                    x, z, intensity_source_dimension, x_first, z_first, intensity_angular_distribution = self.loadASCIIFiles()

                self.progressBarSet(50)

                self.setStatusMessage("Applying new Spatial/Angular Distribution")

                self.progressBarSet(60)

                self.generate_user_defined_distribution_from_srw(beam_out=beam_out,
                                                                 coord_x=x,
                                                                 coord_z=z,
                                                                 intensity=intensity_source_dimension,
                                                                 distribution_type=Distribution.POSITION,
                                                                 kind_of_sampler=self.kind_of_sampler,
                                                                 seed=0 if self.seed==0 else self.seed+1)

                self.progressBarSet(70)

                self.generate_user_defined_distribution_from_srw(beam_out=beam_out,
                                                                 coord_x=x_first,
                                                                 coord_z=z_first,
                                                                 intensity=intensity_angular_distribution,
                                                                 distribution_type=Distribution.DIVERGENCE,
                                                                 kind_of_sampler=self.kind_of_sampler,
                                                                 seed=0 if self.seed==0 else self.seed+2)

                self.setStatusMessage("Plotting Results")

                self.progressBarSet(80)
                self.plot_results(beam_out)

            self.setStatusMessage("")

            if self.compute_power and self.energy_step and total_power:
                additional_parameters = {}

                additional_parameters["total_power"]        = total_power
                additional_parameters["photon_energy_step"] = self.energy_step

                if self.test_mode:
                    additional_parameters["test_mode"] = True

                print("Total Power", total_power)

                beam_out.setScanningData(ShadowBeam.ScanningData("photon_energy", self.energy, "Energy for Power Calculation", "eV", additional_parameters))

            self.send("Beam", beam_out)
        except Exception as exception:
            QtWidgets.QMessageBox.critical(self, "Error", str(exception), QtWidgets.QMessageBox.Ok)

            if self.IS_DEVELOP: raise exception

        self.progressBarFinished()

    def sendNewBeam(self, trigger):
        self.compute_power = False
        self.energy_step = None
        self.test_mode = False

        if trigger and trigger.new_object == True:
            if trigger.has_additional_parameter("seed_increment"):
                self.seed += trigger.get_additional_parameter("seed_increment")

            if trigger.has_additional_parameter("energy_value") and trigger.has_additional_parameter("energy_step"):
                self.compute_power = True
                self.use_harmonic = 1
                self.distribution_source = 0
                self.save_srw_result = 0

                self.energy = trigger.get_additional_parameter("energy_value")
                self.energy_step = trigger.get_additional_parameter("energy_step")
                self.power_step = trigger.get_additional_parameter("power_step")

                self.test_mode = trigger.has_additional_parameter("test_mode") and trigger.get_additional_parameter("test_mode") == True

                self.set_WFUseHarmonic()
                self.set_DistributionSource()
                self.set_SaveFileSRW()

            self.runShadowSource()

    def checkFields(self):
        self.number_of_rays = congruence.checkPositiveNumber(self.number_of_rays, "Number of rays")
        self.seed = congruence.checkPositiveNumber(self.seed, "Seed")

        if self.use_harmonic == 0:
            if self.distribution_source != 0: raise Exception("Harmonic Energy can be computed only for explicit SRW Calculation")

            self.harmonic_number = congruence.checkStrictlyPositiveNumber(self.harmonic_number, "Harmonic Number")
        else:
            self.energy = congruence.checkPositiveNumber(self.energy, "Photon Energy")

        if self.optimize_source > 0:
            self.max_number_of_rejected_rays = congruence.checkPositiveNumber(self.max_number_of_rejected_rays,
                                                                             "Max number of rejected rays")
            congruence.checkFile(self.optimize_file_name)

    def populateFields(self, shadow_src):
        shadow_src.src.NPOINT = self.number_of_rays
        shadow_src.src.ISTAR1 = self.seed
        shadow_src.src.F_OPD = 1
        shadow_src.src.F_SR_TYPE = 0

        shadow_src.src.FGRID = 0
        shadow_src.src.IDO_VX = 0
        shadow_src.src.IDO_VZ = 0
        shadow_src.src.IDO_X_S = 0
        shadow_src.src.IDO_Y_S = 0
        shadow_src.src.IDO_Z_S = 0

        shadow_src.src.FSOUR = 0 # spatial_type (point)
        shadow_src.src.FDISTR = 1 # angular_distribution (flat)

        shadow_src.src.HDIV1 = -1.0e-6
        shadow_src.src.HDIV2 = 1.0e-6
        shadow_src.src.VDIV1 = -1.0e-6
        shadow_src.src.VDIV2 = 1.0e-6

        shadow_src.src.FSOURCE_DEPTH = 1 # OFF

        shadow_src.src.F_COLOR = 1 # single value
        shadow_src.src.F_PHOT = 0 # eV , 1 Angstrom

        shadow_src.src.PH1 = self.energy if self.use_harmonic==1 else self.resonance_energy(harmonic=self.harmonic_number)

        shadow_src.src.F_POLAR = self.polarization

        if self.polarization == 1:
            shadow_src.src.F_COHER = self.coherent_beam
            shadow_src.src.POL_ANGLE = self.phase_diff
            shadow_src.src.POL_DEG = self.polarization_degree

        shadow_src.src.F_OPD = 1
        shadow_src.src.F_BOUND_SOUR = self.optimize_source
        if self.optimize_source > 0:
            shadow_src.src.FILE_BOUND = bytes(congruence.checkFileName(self.optimize_file_name), 'utf-8')
        shadow_src.src.NTOTALPOINT = self.max_number_of_rejected_rays

    # WEIRD MEMORY INITIALIZATION BY FORTRAN. JUST A FIX.
    def fix_Intensity(self, beam_out):
        if self.polarization == 0:
            for index in range(0, len(beam_out._beam.rays)):
                beam_out._beam.rays[index, 15] = 0
                beam_out._beam.rays[index, 16] = 0
                beam_out._beam.rays[index, 17] = 0

    ####################################################################################
    # SRW CALCULATION
    ####################################################################################

    def checkSRWFields(self):

        congruence.checkPositiveNumber(self.Kh, "Horizontal K")
        congruence.checkPositiveNumber(self.Kv, "Vertical K")
        congruence.checkStrictlyPositiveNumber(self.undulator_period, "Period Length")
        congruence.checkStrictlyPositiveNumber(self.number_of_periods, "Number of Periods")

        congruence.checkStrictlyPositiveNumber(self.electron_energy_in_GeV, "Energy")
        congruence.checkPositiveNumber(self.electron_energy_spread, "Energy Spread")
        congruence.checkStrictlyPositiveNumber(self.ring_current, "Ring Current")

        congruence.checkPositiveNumber(self.electron_beam_size_h       , "Horizontal Beam Size")
        congruence.checkPositiveNumber(self.electron_beam_divergence_h , "Vertical Beam Size")
        congruence.checkPositiveNumber(self.electron_beam_size_v       , "Horizontal Beam Divergence")
        congruence.checkPositiveNumber(self.electron_beam_divergence_v , "Vertical Beam Divergence")


        congruence.checkStrictlyPositiveNumber(self.source_dimension_wf_h_slit_gap, "Wavefront Propagation H Slit Gap")
        congruence.checkStrictlyPositiveNumber(self.source_dimension_wf_v_slit_gap, "Wavefront Propagation V Slit Gap")
        congruence.checkStrictlyPositiveNumber(self.source_dimension_wf_h_slit_points, "Wavefront Propagation H Slit Points")
        congruence.checkStrictlyPositiveNumber(self.source_dimension_wf_v_slit_points, "Wavefront Propagation V Slit Points")
        congruence.checkGreaterOrEqualThan(self.source_dimension_wf_distance, self.get_minimum_propagation_distance(),
                                           "Wavefront Propagation Distance", "Minimum Distance out of the Source: " + str(self.get_minimum_propagation_distance()))

        if self.save_srw_result == 1:
            congruence.checkDir(self.source_dimension_srw_file)
            congruence.checkDir(self.angular_distribution_srw_file)

    def get_minimum_propagation_distance(self):
        return round(self.get_source_length()*1.01, 6)

    def get_source_length(self):
        return self.undulator_period*self.number_of_periods

    def magnetic_field_from_K(self):
        Bv = self.Kv * 2 * pi * codata.m_e * codata.c / (codata.e * self.undulator_period)
        Bh = self.Kh * 2 * pi * codata.m_e * codata.c / (codata.e * self.undulator_period)

        return Bv, Bh

    def createUndulator(self):
        #***********Undulator
        By, Bx = self.magnetic_field_from_K() #Peak Vertical field [T]

        phBy = 0 #Initial Phase of the Vertical field component
        sBy = -1 #Symmetry of the Vertical field component vs Longitudinal position
        xcID = 0 #Transverse Coordinates of Undulator Center [m]
        ycID = 0
        zcID = 0 #Longitudinal Coordinate of Undulator Center [m]

        und = SRWLMagFldU([SRWLMagFldH(1, 'h', Bx, phBy, sBy, 1), SRWLMagFldH(1, 'v', By, phBy, sBy, 1)], self.undulator_period, self.number_of_periods) #Planar Undulator
        magFldCnt = SRWLMagFldC([und], array('d', [xcID]), array('d', [ycID]), array('d', [zcID])) #Container of all Field Elements

        return magFldCnt

    def createElectronBeam(self):
        #***********Electron Beam
        elecBeam = SRWLPartBeam()

        elecBeam.partStatMom1.x = 0. #Initial Transverse Coordinates (initial Longitudinal Coordinate will be defined later on) [m]
        elecBeam.partStatMom1.y = 0. #-0.00025
        # Roughly ! check!
        elecBeam.partStatMom1.z = -0.5*self.undulator_period*(self.number_of_periods + 8) #Initial Longitudinal Coordinate (set before the ID)
        elecBeam.partStatMom1.xp = 0. #Initial Relative Transverse Velocities
        elecBeam.partStatMom1.yp = 0.
        elecBeam.partStatMom1.gamma = self.gamma()

        elecBeam.Iavg = self.ring_current #Average Current [A]

        #2nd order statistical moments
        elecBeam.arStatMom2[0] = (self.electron_beam_size_h)**2 #<(x-x0)^2>
        elecBeam.arStatMom2[1] = 0
        elecBeam.arStatMom2[2] = (self.electron_beam_divergence_h)**2 #<(x'-x'0)^2>
        elecBeam.arStatMom2[3] = (self.electron_beam_size_v)**2 #<(y-y0)^2>
        elecBeam.arStatMom2[4] = 0
        elecBeam.arStatMom2[5] = (self.electron_beam_divergence_v)**2 #<(y'-y'0)^2>
        # energy spread
        elecBeam.arStatMom2[10] = (self.electron_energy_spread)**2 #<(E-E0)^2>/E0^2

        return elecBeam

    def createInitialWavefrontMesh(self, elecBeam):
        #****************** Initial Wavefront
        wfr = SRWLWfr() #For intensity distribution at fixed photon energy
        wfr.allocate(1, self.source_dimension_wf_h_slit_points, self.source_dimension_wf_v_slit_points) #Numbers of points vs Photon Energy, Horizontal and Vertical Positions
        wfr.mesh.zStart = self.source_dimension_wf_distance #Longitudinal Position [m] from Center of Straight Section at which SR has to be calculated
        wfr.mesh.eStart = self.energy if self.use_harmonic==1 else self.resonance_energy(harmonic=self.harmonic_number)  #Initial Photon Energy [eV]
        wfr.mesh.eFin = wfr.mesh.eStart #Final Photon Energy [eV]

        wfr.mesh.xStart = -0.5*self.source_dimension_wf_h_slit_gap #Initial Horizontal Position [m]
        wfr.mesh.xFin = -1 * wfr.mesh.xStart #0.00015 #Final Horizontal Position [m]
        wfr.mesh.yStart = -0.5*self.source_dimension_wf_v_slit_gap #Initial Vertical Position [m]
        wfr.mesh.yFin = -1 * wfr.mesh.yStart#0.00015 #Final Vertical Position [m]

        wfr.partBeam = elecBeam

        return wfr

    def createCalculationPrecisionSettings(self):
        #***********Precision Parameters for SR calculation
        meth = 1 #SR calculation method: 0- "manual", 1- "auto-undulator", 2- "auto-wiggler"
        relPrec = 0.01 #relative precision
        zStartInteg = 0 #longitudinal position to start integration (effective if < zEndInteg)
        zEndInteg = 0 #longitudinal position to finish integration (effective if > zStartInteg)
        npTraj = 100000 #Number of points for trajectory calculation
        useTermin = 1 #Use "terminating terms" (i.e. asymptotic expansions at zStartInteg and zEndInteg) or not (1 or 0 respectively)
        arPrecParSpec = [meth, relPrec, zStartInteg, zEndInteg, npTraj, useTermin, 0]

        return arPrecParSpec

    def createBeamlineSourceDimension(self, wfr):
        #***************** Optical Elements and Propagation Parameters

        opDrift = SRWLOptD(-wfr.mesh.zStart)
        ppDrift = [0, 0, 1., 1, 0,
                   self.horizontal_range_modification_factor_at_resizing,
                   self.horizontal_resolution_modification_factor_at_resizing,
                   self.vertical_range_modification_factor_at_resizing,
                   self.vertical_resolution_modification_factor_at_resizing,
                   0, 0, 0]

        return SRWLOptC([opDrift],[ppDrift])

    def transform_srw_array(self, arI, mesh):
        dim_x = mesh.nx
        dim_y = mesh.ny

        x_coordinates = numpy.linspace(mesh.xStart, mesh.xFin, dim_x)
        y_coordinates = numpy.linspace(mesh.yStart, mesh.yFin, dim_y)

        data = numpy.squeeze(arI)
        np_array = data.reshape((dim_y, dim_x))
        np_array = np_array.transpose()

        return x_coordinates, y_coordinates, np_array


    def runSRWCalculation(self):

        self.checkSRWFields()

        magFldCnt = self.createUndulator()
        elecBeam = self.createElectronBeam()
        wfr = self.createInitialWavefrontMesh(elecBeam)
        optBLSouDim = self.createBeamlineSourceDimension(wfr)

        arPrecParSpec = self.createCalculationPrecisionSettings()

        # This is the convergence parameter. Higher is more accurate but slower!!
        # 0.2 is used in the original example. But I think it should be higher. The calculation may then however need too much memory.
        sampFactNxNyForProp = 0.0 #0.6 #sampling factor for adjusting nx, ny (effective if > 0)

        # 1 calculate intensity distribution ME convoluted for dimension size

        arPrecParSpec[6] = sampFactNxNyForProp #sampling factor for adjusting nx, ny (effective if > 0)
        srwl.CalcElecFieldSR(wfr, 0, magFldCnt, arPrecParSpec)

        arI = array('f', [0]*wfr.mesh.nx*wfr.mesh.ny) #"flat" 2D array to take intensity data
        srwl.CalcIntFromElecField(arI, wfr, 6, 1, 3, wfr.mesh.eStart, 0, 0)

        # from radiation at the slit we can calculate Angular Distribution and Power

        x, z, intensity_angular_distribution = self.transform_srw_array(arI, wfr.mesh)

        if self.compute_power:
            if self.power_step > 0:
                total_power = self.power_step
            else:
                dx = (x[1] - x[0])*1e3 # mm for power computations
                dy = (z[1] - z[0])*1e3

                total_power = intensity_angular_distribution.sum() * dx * dy * (1e3 * self.energy_step * codata.e)
        else:
            total_power = None

        distance = wfr.mesh.zStart

        x_first = numpy.arctan(x/distance)
        z_first = numpy.arctan(z/distance)

        wfrAngDist = self.createInitialWavefrontMesh(elecBeam)
        wfrAngDist.mesh.xStart = numpy.arctan(wfr.mesh.xStart/distance)
        wfrAngDist.mesh.xFin   = numpy.arctan(wfr.mesh.xFin  /distance)
        wfrAngDist.mesh.yStart = numpy.arctan(wfr.mesh.yStart/distance)
        wfrAngDist.mesh.yFin   = numpy.arctan(wfr.mesh.yFin  /distance)

        if self.save_srw_result == 1: srwl_uti_save_intens_ascii(arI, wfrAngDist.mesh, self.angular_distribution_srw_file)

        # for source dimension, back propagation to the source central position

        srwl.PropagElecField(wfr, optBLSouDim)

        arI = array('f', [0]*wfr.mesh.nx*wfr.mesh.ny) #"flat" 2D array to take intensity data
        srwl.CalcIntFromElecField(arI, wfr, 6, 1, 3, wfr.mesh.eStart, 0, 0)

        if self.save_srw_result == 1: srwl_uti_save_intens_ascii(arI, wfr.mesh, self.source_dimension_srw_file)

        x, z, intensity_source_dimension = self.transform_srw_array(arI, wfr.mesh)

        # SWITCH FROM SRW METERS TO SHADOWOUI U.M.
        x /= self.workspace_units_to_m
        z /= self.workspace_units_to_m

        return x, z, intensity_source_dimension, x_first, z_first, intensity_angular_distribution, total_power


    def generate_user_defined_distribution_from_srw(self,
                                                    beam_out,
                                                    coord_x,
                                                    coord_z,
                                                    intensity,
                                                    distribution_type=Distribution.POSITION,
                                                    kind_of_sampler=0,
                                                    seed=0):
        if kind_of_sampler == 0:
            pdf = numpy.abs(intensity/numpy.max(intensity))
            pdf /= pdf.sum()
    
            distribution = CustomDistribution(pdf, seed=seed)
    
            sampled = distribution(len(beam_out._beam.rays))
    
            min_value_x = numpy.min(coord_x)
            step_x = numpy.abs(coord_x[1]-coord_x[0])
            min_value_z = numpy.min(coord_z)
            step_z = numpy.abs(coord_z[1]-coord_z[0])
    
            if distribution_type == Distribution.POSITION:
                beam_out._beam.rays[:, 0] = min_value_x + sampled[0, :]*step_x
                beam_out._beam.rays[:, 2] = min_value_z + sampled[1, :]*step_z
    
            elif distribution_type == Distribution.DIVERGENCE:
                alpha_x = min_value_x + sampled[0, :]*step_x
                alpha_z = min_value_z + sampled[1, :]*step_z
    
                beam_out._beam.rays[:, 3] =  numpy.cos(alpha_z)*numpy.sin(alpha_x)
                beam_out._beam.rays[:, 4] =  numpy.cos(alpha_z)*numpy.cos(alpha_x)
                beam_out._beam.rays[:, 5] =  numpy.sin(alpha_z)
        elif kind_of_sampler == 1:
            min_x = numpy.min(coord_x)
            max_x = numpy.max(coord_x)
            delta_x = max_x - min_x
    
            min_z = numpy.min(coord_z)
            max_z = numpy.max(coord_z)
            delta_z = max_z - min_z
    
            dim_x = len(coord_x)
            dim_z = len(coord_z)
    
            grid = Grid2D((dim_x, dim_z))
            grid[..., ...] = intensity.tolist()
    
            d = Distribution2D(distribution_from_grid(grid, dim_x, dim_z), (0, 0), (dim_x, dim_z))
    
            samples = d.get_samples(len(beam_out._beam.rays), seed)
    
            if distribution_type == Distribution.POSITION:
                beam_out._beam.rays[:, 0] = min_x + samples[:, 0]*delta_x
                beam_out._beam.rays[:, 2] = min_z + samples[:, 1]*delta_z
    
            elif distribution_type == Distribution.DIVERGENCE:
                alpha_x = min_x + samples[:, 0]*delta_x
                alpha_z = min_z + samples[:, 1]*delta_z
    
                beam_out._beam.rays[:, 3] =  numpy.cos(alpha_z)*numpy.sin(alpha_x)
                beam_out._beam.rays[:, 4] =  numpy.cos(alpha_z)*numpy.cos(alpha_x)
                beam_out._beam.rays[:, 5] =  numpy.sin(alpha_z)
        else:
            raise ValueError("Sampler not recognized")
        
    def gamma(self):
        return 1e9*self.electron_energy_in_GeV / (codata.m_e *  codata.c**2 / codata.e)

    def resonance_energy(self, theta_x=0.0, theta_z=0.0, harmonic=1):
        gamma = self.gamma()

        wavelength = (self.undulator_period / (2.0*gamma **2)) * \
                     (1 + self.Kv**2 / 2.0 + self.Kh**2 / 2.0 + \
                      gamma**2 * (theta_x**2 + theta_z ** 2))

        wavelength /= harmonic

        return m2ev/wavelength

    ####################################################################################
    # SRW FILES
    ####################################################################################

    def checkSRWFilesFields(self):
        congruence.checkFile(self.source_dimension_srw_file)
        congruence.checkFile(self.angular_distribution_srw_file)

    def loadSRWFiles(self):

        self.checkSRWFilesFields()

        x, z, intensity_source_dimension = self.loadNumpyFormat(self.source_dimension_srw_file)
        x_first, z_first, intensity_angular_distribution = self.loadNumpyFormat(self.angular_distribution_srw_file)


        # SWITCH FROM SRW METERS TO SHADOWOUI U.M.
        x = x/self.workspace_units_to_m
        z = z/self.workspace_units_to_m

        return x, z, intensity_source_dimension, x_first, z_first, intensity_angular_distribution


    def file_load(self, _fname, _read_labels=1):
        nLinesHead = 11
        hlp = []

        with open(_fname,'r') as f:
            for i in range(nLinesHead):
                hlp.append(f.readline())

        ne, nx, ny = [int(hlp[i].replace('#','').split()[0]) for i in [3,6,9]]
        ns = 1
        testStr = hlp[nLinesHead - 1]
        if testStr[0] == '#':
            ns = int(testStr.replace('#','').split()[0])

        e0,e1,x0,x1,y0,y1 = [float(hlp[i].replace('#','').split()[0]) for i in [1,2,4,5,7,8]]

        data = numpy.squeeze(numpy.loadtxt(_fname, dtype=numpy.float64)) #get data from file (C-aligned flat)

        allrange = e0, e1, ne, x0, x1, nx, y0, y1, ny

        arLabels = ['Photon Energy', 'Horizontal Position', 'Vertical Position', 'Intensity']
        arUnits = ['eV', 'm', 'm', 'ph/s/.1%bw/mm^2']

        if _read_labels:

            arTokens = hlp[0].split(' [')
            arLabels[3] = arTokens[0].replace('#','')
            arUnits[3] = '';
            if len(arTokens) > 1:
                arUnits[3] = arTokens[1].split('] ')[0]

            for i in range(3):
                arTokens = hlp[i*3 + 1].split()
                nTokens = len(arTokens)
                nTokensLabel = nTokens - 3
                nTokensLabel_mi_1 = nTokensLabel - 1
                strLabel = ''
                for j in range(nTokensLabel):
                    strLabel += arTokens[j + 2]
                    if j < nTokensLabel_mi_1: strLabel += ' '
                arLabels[i] = strLabel
                arUnits[i] = arTokens[nTokens - 1].replace('[','').replace(']','')

        return data, None, allrange, arLabels, arUnits

    def loadNumpyFormat(self, filename):
        data, dump, allrange, arLabels, arUnits = self.file_load(filename)

        dim_x = allrange[5]
        dim_y = allrange[8]
        np_array = data.reshape((dim_y, dim_x))
        np_array = np_array.transpose()
        x_coordinates = numpy.linspace(allrange[3], allrange[4], dim_x)
        y_coordinates = numpy.linspace(allrange[6], allrange[7], dim_y)

        return x_coordinates, y_coordinates, np_array

    ####################################################################################
    # ASCII FILES
    ####################################################################################

    def checkASCIIFilesFields(self):
        congruence.checkFile(self.x_positions_file)
        congruence.checkFile(self.z_positions_file)
        congruence.checkFile(self.x_divergences_file)
        congruence.checkFile(self.z_divergences_file)

        self.x_positions_factor = float(self.x_positions_factor)
        self.z_positions_factor = float(self.z_positions_factor)
        self.x_divergences_factor = float(self.x_divergences_factor)
        self.z_divergences_factor = float(self.z_divergences_factor)

        congruence.checkStrictlyPositiveNumber(self.x_positions_factor, "X Position Units to Workspace Units")
        congruence.checkStrictlyPositiveNumber(self.z_positions_factor, "Z Position Units to Workspace Units")
        congruence.checkStrictlyPositiveNumber(self.x_divergences_factor, "X Divergence Units to rad")
        congruence.checkStrictlyPositiveNumber(self.z_divergences_factor, "Z Divergence Units to rad")

    def loadASCIIFiles(self):
        self.checkASCIIFilesFields()

        x_positions = self.extract_distribution_from_file(distribution_file_name=self.x_positions_file)
        z_positions = self.extract_distribution_from_file(distribution_file_name=self.z_positions_file)

        x_positions[:, 0] *= self.x_positions_factor
        z_positions[:, 0] *= self.z_positions_factor

        x_divergences = self.extract_distribution_from_file(distribution_file_name=self.x_divergences_file)
        z_divergences = self.extract_distribution_from_file(distribution_file_name=self.z_divergences_file)

        x_divergences[:, 0] *= self.x_divergences_factor
        z_divergences[:, 0] *= self.z_divergences_factor

        x, z, intensity_source_dimension = self.combine_distributions(x_positions, z_positions)
        x_first, z_first, intensity_angular_distribution = self.combine_distributions(x_divergences, z_divergences)

        return x, z, intensity_source_dimension, x_first, z_first, intensity_angular_distribution


    def extract_distribution_from_file(self, distribution_file_name):
        distribution = []

        try:
            distribution_file = open(distribution_file_name, "r")

            rows = distribution_file.readlines()

            for index in range(0, len(rows)):
                row = rows[index]

                if not row.strip() == "":
                    values = row.split()

                    if not len(values) == 2: raise Exception("Malformed file, must be: <value> <spaces> <frequency>")

                    value = float(values[0].strip())
                    frequency = float(values[1].strip())

                    distribution.append([value, frequency])

        except Exception as err:
            raise Exception("Problems reading distribution file: {0}".format(err))
        except:
            raise Exception("Unexpected error reading distribution file: ", sys.exc_info()[0])

        return numpy.array(distribution)

    def combine_distributions(self, distribution_x, distribution_y):

        coord_x = distribution_x[:, 0]
        coord_y = distribution_y[:, 0]

        intensity_x = repmat(distribution_x[:, 1], len(coord_y), 1).transpose()
        intensity_y = repmat(distribution_y[:, 1], len(coord_x), 1)

        if self.combine_strategy == 0:
            convoluted_intensity = numpy.sqrt(intensity_x*intensity_y)
        elif self.combine_strategy == 1:
            convoluted_intensity = numpy.sqrt(intensity_x**2 + intensity_y**2)
        elif self.combine_strategy == 2:
            convoluted_intensity = convolve2d(intensity_x, intensity_y, boundary='fill', mode='same', fillvalue=0)
        elif self.combine_strategy == 3:
            convoluted_intensity = 0.5*(intensity_x + intensity_y)

        return coord_x, coord_y, convoluted_intensity


